#-----------------------------------------------------------------------------
#  Copyright (c) 2016 Eclektek LLC. All rights reserved.
#
#
#  Developed by: William Carter
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and  
#  associated documentation files (the "Software"), to deal with the Software without restriction,  
#  including without limitation the rights to use, copy, modify, merge, publish, sublicense and/or    
#  distribute copies of the Software, and to permit persons to whom the Software is furnished to do so,  
#  subject to the following conditions:
#
#      Redistributions of source code must retain the above copyright notice, this list of conditions 
#      and the following disclaimers.
#      Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
#      and the following disclaimers in the documentation and/or other materials provided with the distribution.
#      Neither the names of William Carter, Eclektek LLC, nor the names of its 
#      contributors may be used to endorse or promote products derived from this Software without 
#      specific prior written permission from Eclektek LLC.
#      This Software is provided free of charge and Eclektek LLC reserves the right to sell copies of the 
#      Software in it original form. Anyone sub-licensing this Software will provide it free of charge.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
#  LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
#  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
#  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
#  SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
#-----------------------------------------------------------------------------
#  History :                                                                       
#                                                                                                                  
#  Date      By    Version  Comments                                         
#  --------  ----  -------  -----------------------------------------------------
#  09/24/16  W.C.  1.00     Created                                     
#
#-----------------------------------------------------------------------------
# Tester.py was written using Python 2.7 on Windows 7
# Tester.py requires the following non-standard Library modules:
#   pyserial-2.6
#   
#-----------------------------------------------------------------------------
# Tester.py communicates with the Differential PMOD CHallenge Testbed HDL component
# to setup, execute and acquire test results using a USB UART
#
# Tester only functions with a Nexys Video or Genesys2 board when configured with one of the
# Differential PMOD Challenge toplevel builds.
#
# Command line Useage: <Python Path>python Tester.py [-p port] [-l off]
#  where <Python Path> is the location of your Python version 2.7 or less installation
#  [port] is the name of the USB UART port connected to your FPGA board UART connector
#  [off] disables data logging ( anything else or absense of the -l options enables data logging
#
#   Windows example:  C:\Python27\python27 Tester.py -p COM9 -l off
#   Linux example:    python Tester.py -p /dev/ttyUSB0 
#
# Tester first checks that the specified UART port is available. If not it exits. Tester then
# tries to write a value to UART_RegAccess Register 0 and read it back. If unsuccessful it exits.
# If the UART_RegAccess test succeeds a menu is displayed in the terminal window or DOS box.
# Tester Commands are:
#   'Q' - Quit Tester
#   'N' - Set the Si570 to a new output frequency. Values are entered in Hz as integers ( no fractional part )
#   'M' - Display the current Si570 frequency measurement. Displayed in Mhz. 
#   'S' - Sweep test. You will be prompted to provide a lower and upper bound as well as an delta all in
#         Hz. Tests will be run at all frequencies from the lower through the upper bound in delta Hz increments.
#   'R' - Run a test at the current Si570 frequency
#   'E' - Enter EYE mode. Data source is run until a 'T' command ends it. This allows using the MONITOR_PMOD
#         data signals to evaluate the "eye" of each data bit ( delayed if the idelay option is enabled ) on
#         an oscilloscope
#   'Z' - Display registers R1 though R11
#   'C' - Clear all status registers
#   'T' - Force the Testbed state machine into the IDLE state and reset the Source and Sink DUT logic. 
#         Required to exit 'eye' mode. 
#   'D' - Set iDELAY values. You will be prompted to enter 4 values from 0 to 31 to set each of the iDELAY
#         delays ( only does anything if the if the idelay option is enabled in the FPGA build). 
#         Delays from 0 to ~ 2.5 ns are possible.
#   'X' - Displays the current FPGA Temperature sensor reading in Celsius degrees
#
# Tester writes a uniquely timestamped log file, unless disabled by the '-l off' command line option, in the directory
# where Tester.py is located. A new file is created each time that Tester is run.
#
# Typically, you might want to run the 'S' command to run through a set of frequencies. Be aware that it might
# be that successful runs occur at particular frequencies related to your sampling clock frequency. It is likely
# that the highest Si570 frequency that will produce a successful test run is found by using the 'N' command to
# set a particular frequency and then a 'R' command to perform a test run at that frequency. Of course, if you want
# there's nothing stopping you from running a Sweep test from  10.5 - 100 MHz in 250 KHz increments.
#  
# A successful test will end with R4 = 0x000000100000 = 1048576 words written to the testbed by the Data Sink without
# an error. R5 displays the total elapsed time that the testbed state machine was in the RUN state and is the time from when
# the first word was sent to the Data Source to when the 1048576th word was received from the Data Sink. The
# elapsed time in seconds is R5/100,000,000. The 32-bit Word data rate is 1048576/elapsed time = 1048576*100000000/R5 
#
# The 'X' command works even if there is nothing connected to the FMC connector. You can verify the fan control by
# configuring the Genesys2, entering "eye" mode, waiting for the fan to come on and executing the 'X' command as soon 
# as the fan comes on and until it goes off again.
# 
#-----------------------------------------------------------------------------
# NOTES:
#
#   - in this version of Tester not all status register fields are extracted and displayed
#   - on some FPGA boards the UART is powered by the USB hub or PC and will be enumerated and available whenever a
#     cable connects the FPGA UART connector to a PC or PC connected hub, even if the FPGA board power switch is off.
#   - in Windows 7 the appropriate UART port can be found using the Windows Device Manager and is typically COMxx. In
#     Linux typing 'dmesg | grep tty' in a terminal will display all of the USB devices that have been enumerated. In either OS
#     comparing what's available when the FPGA UART is not attached to what's available when it is attached helps 
#     identify the correct port in case of confusion.
#   - the lowest output frequency supported by the Si570 on the Xilinx HW-FMC-105-DEBUG board is 10 MHz, though the
#     lowest I've been able to set is around 10.5 MHZ. See the Si570 datasheet from Silicon Labs for information.
# 
#-----------------------------------------------------------------------------
# ToDo:
#
#
#-----------------------------------------------------------------------------
# KNOWN ISSUES:
#
#   - After updating the Si570 configuration registers two other registers must be manipulated within 10 ms in order
#     to obtain the new frequency. Depending on other OS activity this may not happen and occasionaly a frequency change
#     will be wrong. A second attempt will generally succeed.
#
#-----------------------------------------------------------------------------

# Modules
import serial
import time
import math
import sys

# Some definitions

# FPGA UART_RegAccess Register Assignments
test_readback_reg         = '0'
test_options_reg          = '1'
debug_status_reg          = '2'
test_status_reg           = '3'
test_data_count_reg       = '4'
elapsed_time_reg          = '5'
sink_lfsr_data_reg        = '6'
sink_data_reg             = '7'
test_control_reg          = '8'
test_idelay_reg           = '9'
xadc_write_reg            = 'A'
xadc_read_reg             = 'B'
freq_measure_reg          = 'C'
si570_config_read_reg     = 'D'
si570_config_write_reg    = 'E'
si570_i2c_command_reg     = 'F'

# Control and Status bit field assignments
## Reg1
use_genesys2_mask         = 0x800000000000
use_differential_io_mask  = 0x400000000000
use_idelay_mask           = 0x200000000000 
use_ddr_mask              = 0x100000000000 
interface_type_mask       = 0x00FF00000000
test_version_mask         = 0x00000000FFFF
## Reg2
test_sm_states_mask       = 0xF00000000000
test_ended_flag_mask      = 0x000000000001
## Reg3
error_detected_mask       = 0x800000000000
timeout_error_mask        = 0x400000000000
sink_pdata_mask           = 0x00000000000F
## Reg4
sink_data_count_mask      = 0x000000FFFFFF
## Reg5
test_elapsed_time_mask    = 0x0000FFFFFFFF
## Reg6
sink_lfsr_data_mask       = 0x0000FFFFFFFF
## Reg7
sink_data_mask            = 0x0000FFFFFFFF
## Reg8
ctrl_start_bit_mask       = 0x000000000001
ctrl_eyemode_bit_mask     = 0x000000000002
ctrl_dut_reset_bit_mask   = 0x000000000004
ctrl_smreset_bit_mask     = 0x000000000008
## Reg9
d0_idelay_mask            = 0x00000000001F
d1_idelay_mask            = 0x000000001F00
d2_idelay_mask            = 0x0000001F0000
d3_idelay_mask            = 0x00001F000000
## Reg10
xadc_wr_bit_mask          = 0x800000000000
xadc_address_mask         = 0x000001FF0000
xadc_write_data           = 0x00000000FFFF
## Reg11
xadc_temp_ot_alarm_mask   = 0x800000000000
xadc_alarm_mask           = 0x7F0000000000
xadc_read_data            = 0x00000000FFFF

# NOTE: in some cases such as R4, R5, R6 and R7 unused bits are 0
#       so the mask values is mainly provided for informational purposes

# Reg15 FPGA I2C Master interface command set
cmd_read_config       = '0'
cmd_write_config      = '1'
cmd_set_dco_freeze    = '2'

# Si570 HS_DIV mapping
hsdiv_map  = {'hsdiv_four':   0, 
              'hsdiv_five':   1,
              'hsdiv_six':    2,
              'hsdiv_seven':  3,
              'hsdiv_bad0':   4,
              'hsdiv_nine':   5,
              'hsdiv_bad1':   6,
              'hsdiv_eleven': 7}
hsdiv_keys = hsdiv_map.keys()

hsdiv_bits_to_val = [4, 5, 6, 7, 20, 9, 30, 11] # only values < 12 are valid

Fref = 100.0; # Sysclk in MHz
si570_dco_max  = 5670000000
si570_dco_min  = 4850000000
si570_dco_target   = 5260000000
si570_fxtal     = 114285000.0
config_default  = 1935889486750 # 0x01C2BC01539E
config_default_sl  = format(config_default, 'x')  # get a long hex value without the extra chars
config_default_s = config_default_sl.upper()
freq_default    = 156250000.0
config_200mhz_s = '60C310017C5F'
two_twenty_eight_pwr = 268435456
two_thirty_seven_pwr = 137438953472
two_thirty_eight_pwr = 274877906944
two_fourty_five_pwr = 35184372088832
two_fourty_four_pwr = 17592186044416
two_to_eight_pwr        = 256
two_to_sixteen_pwr      = 65536 
two_to_twenty_four_pwr  = 16777216 
two_to_thirty_two_pwr  = 4294967296 
channel_idelays = [31, 31, 31, 31] 


# Test Setup from the FPGA
setup = {'board':         'GENESYS2', 
         'iostandard':    'TMDS33',
         'idelay_status': 'NO IDELAY',
         'iddr_status':   'NO IDDR',
         'interface':     'SDR',
         'version':       '1'
        } 


# command line options
uart_port = 'COM9'
save_log = 1

# create a database for configuration parameters using a dictionary of lists
freq_params = {'fdefault':[156250000,'01C2BC01539E',4,8,11744138142], 'f200000000':[200000000,'60C310017C5F',4,8,13153434719]}
# parameter list entry indices
fparams_freq = 0  
fparams_config = 1
fparams_hsdiv = 2 
fparams_n = 3     
fparams_fref = 4  

# Functions

def get_results():
  t = [0,0]
  TestData = 'R 4\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  x = int(ResponseData[4:18],16)
  t[0] = x
  TestData = 'R 5\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  x = int(ResponseData[4:18],16)
  t[1] = x
  return t
  
  
def get_test_setup(s):
  TestData = 'R 1\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  n = int(ResponseData[4:18],16)
  if n & use_genesys2_mask == 0:
    s['board'] = 'NEXYSVIDEO'
  else:
    s['board'] = 'GENESYS2'
  if n & use_differential_io_mask == 0:
    s['iostandard'] = 'LVCMOS33'
  else:
    s['iostandard'] = 'TMDS33'
  if n & use_idelay_mask == 0:
    s['idelay_status'] = 'NO IDELAY'
  else:
    s['iostandard'] = 'IDELAY'
  if n & use_ddr_mask == 0:
    s['iddr_status'] = 'NO IDDR'
  else:
    s['iddr_status'] = 'IDDR'
  if n & interface_type_mask == 0:
    s['interface'] = 'SDR'
  else:
    s['interface'] = 'Bi-phaseL'
  s['version'] = str(int(n & test_version_mask))
  return s

      
# get command-line options
def get_options(argv):
  clo= {}
  while argv:
    if argv[0][0] == '-':
      clo[argv[0]] = argv[1]
      argv = argv[2:]
    else:
      argv = argv[1:]
  return clo
  

# Set channel input delays
def set_delays(d):
  print 'Current Delays:' 
  i = 0
  e = [31,31,31,31] 
  print d    
  for x in d:
    c = raw_input('Enter CH' + str(i) + ' Delay (0 to 31) => ')
    print c
    if int(c) < 31:
      e[i] = int(c)
    i = i + 1
  print d 
  v = e[0] + e[1]*two_to_eight_pwr + e[2]*two_to_sixteen_pwr+ e[3]*two_to_twenty_four_pwr
  s = format(v,'012x')
  TestData = 'W ' + test_idelay_reg + ' ' + s.upper() + '\x0D\x0A'
  ser.write(TestData)
  time.sleep(.2)              # delay 
  TestData = 'R ' + test_idelay_reg + '\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reg9 = ', ResponseData[4:18]   
  return e  

# Temporary xadc function
# Write Reg11  
def write_xadc():
  s = raw_input('Enter Reg11 data => ')
  print s
  TestData = 'W ' + xadc_write_reg + ' ' + s.upper() + '\x0D\x0A'
  ser.write(TestData)
 
# Display FPGA core temperature in Celsius degrees
def get_xadc_temp():
  TestData = 'W ' + xadc_write_reg + ' 0' + '\x0D\x0A'
  ser.write(TestData)
  time.sleep(.2)              # delay 
  TestData = 'R ' + xadc_read_reg + '\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  n = int(ResponseData[4:18], 16)
  t = float(n)*2015.0/(16.0*16384.0) - 273.0
  print 'FPGA Core Temperature = ', t   
    

# function to clear all status registers
def clear_status():
  TestData = 'W 1 0\x0D\x0A'
  ser.write(TestData)
  time.sleep(.2)              # delay 
  TestData = 'W 2 0\x0D\x0A'
  ser.write(TestData)
  time.sleep(.2)              # delay 
  TestData = 'W 3 0\x0D\x0A'
  ser.write(TestData)
  time.sleep(.2)              # delay 
  TestData = 'W 4 0\x0D\x0A'
  ser.write(TestData)
  time.sleep(.2)              # delay 
  TestData = 'W 5 0\x0D\x0A'
  ser.write(TestData)
  time.sleep(.2)              # delay 
  TestData = 'W 6 0\x0D\x0A'
  ser.write(TestData)
  time.sleep(.2)              # delay 
  TestData = 'W 7 0\x0D\x0A'
  ser.write(TestData)
  time.sleep(.2)              # delay 

# function to read all status registers
def read_status():
  TestData = 'R 1\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reg1 = ', ResponseData[4:18]
  if save_log == 1:
    print >> log,'Reg1 = ', ResponseData[4:18]
  TestData = 'R 2\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reg2 = ', ResponseData[4:18]
  if save_log == 1:
    print >> log,'Reg2 = ', ResponseData[4:18]
  TestData = 'R 3\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reg3 = ', ResponseData[4:18]
  if save_log == 1:
    print >> log,'Reg3 = ', ResponseData[4:18]
  TestData = 'R 4\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reg4 = ', ResponseData[4:18]
  if save_log == 1:
    print >> log,'Reg4 = ', ResponseData[4:18]
  TestData = 'R 5\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reg5 = ', ResponseData[4:18]
  if save_log == 1:
    print >> log,'Reg5 = ', ResponseData[4:18]
  TestData = 'R 6\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reg6 = ', ResponseData[4:18]
  if save_log == 1:
    print >> log,'Reg6 = ', ResponseData[4:18]
  TestData = 'R 7\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reg7 = ', ResponseData[4:18]
  if save_log == 1:
    print >> log,'Reg7 = ', ResponseData[4:18]
  TestData = 'R 8\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reg8 = ', ResponseData[4:18]
  if save_log == 1:
    print >> log,'Reg8 = ', ResponseData[4:18]
  TestData = 'R 9\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reg9 = ', ResponseData[4:18]
  if save_log == 1:
    print >> log,'Reg9 = ', ResponseData[4:18]
  TestData = 'R A\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reg10 = ', ResponseData[4:18]
  if save_log == 1:
    print >> log,'Reg10 = ', ResponseData[4:18]
  TestData = 'R B\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reg11 = ', ResponseData[4:18]
  if save_log == 1:
    print >> log,'Reg11 = ', ResponseData[4:18]
  

# function to force the state machine into IDLE
def terminate_test():
  TestData = 'W 8 C\x0D\x0A'  # reset the data sink and data source DUTs, force IDLE state
  ser.write(TestData)
  time.sleep(.2)              # delay 
  TestData = 'W 8 0\x0D\x0A'
  time.sleep(.2)              # delay 
  
  
# function to execute a test run
def run_test():
  terminate_test()
  clear_status()              # Reg2(0) must be 0 before we can run another test
  TestData = 'W 8 1\x0D\x0A'  # assert start_stb
  ser.write(TestData)
  time.sleep(.2)              # delay 
  TestData = 'W 8 0\x0D\x0A'  # de-assert start_stb
  ser.write(TestData)
  
# function to enter EYE mode
def run_eye():
  terminate_test()
  clear_status()              # Reg2(0) must be 0 before we can run another test
  TestData = 'W 8 3\x0D\x0A'  # assert start_stb and 
  ser.write(TestData)
  time.sleep(.2)              # delay 
  TestData = 'W 8 2\x0D\x0A'  # de-assert start_stb
  ser.write(TestData)
  
  
# function to read RegC and print the calculated Si570 frequency
def show_freq():
  # Get the default vclk period measurement
  TestData = 'R ' + freq_measure_reg + '\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  # print out measured frequency
  p = int(ResponseData[4:18],16)  #convert string of hex digits to decimal
  freq = Fref*1048576/(2*p)
  print 'Measured Frequency (Mhz) = ',freq
  if save_log == 1:
    print >> log, 'Measured Frequency (Mhz) = ',freq  

# function to read RegD and return a number  
def read_config():
  # Get default configuration
  TestData = 'W ' + si570_i2c_command_reg + ' ' + cmd_read_config + '\x0D\x0A' 
  ser.write(TestData)
  time.sleep(.2)              # delay 
  TestData = 'R ' + si570_config_read_reg + '\x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Current Configuration = ', ResponseData[4:18]
  if save_log == 1:
    print >> log, 'Current Configuration = ', ResponseData[4:18]
  return int(ResponseData[4:18], 16)  

# Given a configuration string change the Si570 frequency  
def write_config(cstr):  
  TestData = 'W ' + si570_config_write_reg + ' '+ cstr + '\x0D\x0A'
  ser.write(TestData)
  # Set the DCO Freeze bit
  TestData = 'W ' + si570_i2c_command_reg + ' ' + cmd_set_dco_freeze + '\x0D\x0A' 
  ser.write(TestData)
  time.sleep(.2)              # delay 
  
  # Write the new configuration
  TestData = 'W ' + si570_i2c_command_reg + ' ' + cmd_write_config + '\x0D\x0A' 
  ser.write(TestData)
  
  # Add a short delay in case a read_config() is called immediately
  # after this one returns else we may read the previouse UART Rx data....
  time.sleep(1) 

# function to extract Si570 parameters from a configuration string
def extract_params(cfgnum, fout):
  #extract the oscillator parameters to numbers
  print 'Extracting parameters from configuration data...'
  n = int(cfgnum, 16)
  hs_div_bits = n >> 44
  hs_div = hsdiv_bits_to_val[hs_div_bits]
  print 'HighSpeedDivider = ',hs_div
  n1_bits = (n & 0x1FC000000000) >> 38
  n1_val = n1_bits + 1
  n1_val = n1_bits + 1
  print 'N1 = ', n1_val
  fref_bits = n & 0x003FFFFFFFFF
  Fdco = fout * n1_val * hs_div
  print 'Fdco = ', Fdco
  Frefq = fref_bits / two_twenty_eight_pwr
  print 'Fref = ' , Frefq 
  Fxtal = Fdco/Frefq
  print 'Fxtal = ', Fxtal

# function to create new freq_params lists based on the default
# input is a float or int
# output is a list compatible with freq_params
def create_params(newfreq):
  p = [0,0,0,0,0]
  m = si570_dco_target//(newfreq) # target hsdiv*n
  # find m/hsdiv where the fractional part is closest to an integer
  # note: max hsdiv and lowest N1 is best power dissipation.... 
  if m > 4:
    x = math.fmod(m,4)
    r = math.fabs(x - int(x))
    hsdiv = 4
    y = 0
    s = r

  if m > 5:
    x = math.fmod(m,5)
    r = math.fabs(x - int(x))
    if r > s:
      s = r
      hsdiv = 5 
      y = 1
      
   
  if m > 6:
    x = math.fmod(m,6)
    r = math.fabs(x - int(x))
    if r > s:
      s = r
      hsdiv = 6 
      y = 2

  if m > 7:
    x = math.fmod(m,7)
    r = math.fabs(x - int(x))
    if r > s:
      s = r
      hsdiv = 7 
      y = 3

  if m > 9:
    x = math.fmod(m,9)
    r = math.fabs(x - int(x))
    if r > s:
      s = r
      hsdiv = 9 
      y = 5

  if m > 11:
    x = math.fmod(m,1)
    r = math.fabs(x - int(x))
    if r > s:
      s = r
      hsdiv = 11
      y = 7
       
  n = int(m//hsdiv) 
  
  if n % 2 > 0: # N has to be either 1 or a even number between 2 and 128
    if m > 3:
      m = m - 1
      n = n + 1  
    else:
      m = m + 1
      n = n - 1  
          
  dco = newfreq*hsdiv*n
  
  if dco > si570_dco_max:
    n = n - 1
    print 'Error High Fcdo', dco 
     
  if dco < si570_dco_min:
    n = n + 1
    print 'Error Low Fcdo', dco
    
  fref = round(dco*two_twenty_eight_pwr/114284868.227)  # 114.285 MHz = Fxtal
  v = int((y*two_fourty_five_pwr)+((n-1)*two_thirty_eight_pwr)+int(fref))
  s = format(v,'012x')
  print m, hsdiv, n, dco, v
  p[fparams_freq] = newfreq
  p[fparams_config] = s.upper()
  p[fparams_hsdiv] = hsdiv
  p[fparams_n] = int(n)
  p[fparams_fref] = int(fref)
  return p
  
def print_cmds():
  print '\nCommands: Q N M S R E Z C T D X'  
  
 
# Begin by checking for command line options
#   -p 'USB PORT NAME'  - USB PORT NAME examples 'COM11' , 'USB1'
#   -l  'off'           - 'off' disable writing log file, anything else is enable (default) 
cmd_line_options = get_options(sys.argv)
if cmd_line_options.has_key('-p') == 1:
  uart_port = cmd_line_options['-p']
  uart_port = uart_port
  print 'Using UART PORT: ' + uart_port
  
if cmd_line_options.has_key('-l') == 1:
  if cmd_line_options['-l'].upper() == 'OFF':
    save_log = 0
    print 'Test Logging is Disabled'

else:
  save_log = 1
  print 'Test Logging is Enabled'
      
  

# Start by trying to open the specified UART port
try:
  ser = serial.Serial(port=uart_port)
except:  
  uart_port_is_open = False
  print uart_port + ' is not available'
else:
  uart_port_is_open = True
  ser.baudrate = 921600 
  # select FIVEBITS, SIXBITS, SEVENBITS or EIGHTBITS of data
  #ser.bytesize = serial.FIVEBITS
  #ser.bytesize = serial.SIXBITS
  #ser.bytesize = serial.SEVENBITS
  ser.bytesize = serial.EIGHTBITS
  
  # select STOPBITS_ONE, STOPBIT_TWO or STOPBITS_ONE_POINT_FIVE
  ser.stopbits = serial.STOPBITS_ONE
  #ser.stopbits = serial.STOPBIT_TWO
  #ser.stopbits = serial.STOPBITS_ONE_POINT_FIVE
  
  # select PARITY_NONE, PARITY_EVEN or PARITY_ODD
  ser.parity = serial.PARITY_NONE
  #ser.parity = serial.PARITY_EVEN
  #ser.parity = serial.PARITY_ODD
  
  ser.xonxoff = 0
  ser.rtscts = 0
  # IF you don't set the timeout for reads the code will hang
  ser.timeout = 0.5
  
  #ser.parity = serial.PARITY_EVEN
  
# only continue if the specified UART connection exists  
if uart_port_is_open == True:
    
  # Test R0 to see if we can communicate with the FPGA
  print 'Writing R0 with 0x112233445566 to test interface'
  TestData = 'W ' + test_readback_reg + ' 112233445566\x0D\x0A' # write command followed by CRLF
  ser.write(TestData)
  TestData = 'R ' + test_readback_reg + ' \x0D\x0A'
  ser.write(TestData)
  ResponseData = ser.read(19)
  print 'Reading R0: ',ResponseData
  if ResponseData[:18] == '0x0 0x112233445566':
    print 'Register write/read Test Successful'
    has_connection = True
  else:  
    print 'Register write/read Test Failed!'
    has_connection = False
  
  # After the initial tests we have the main menu
  # but only if we were able to read back R0 successfully
  if has_connection == True:
    setup = get_test_setup(setup)
    print 'Test Setup: ', 
    print setup['board'] + ' ',
    print setup['iostandard'] + ' ',
    print setup['idelay_status'] + ' ',
    print setup['iddr_status'] + ' ', 
    print setup['interface'] + ' ', 
    print 'FPGA Version = ' + setup['version'],
    print
  
    # open a log file if logging is enabled
    if save_log == 1:
      tstr = time.ctime(time.time())
      tstr1 = tstr.replace(' ', '_')
      tstr1 = tstr.replace(':', '')
      tstr1 = setup['board'] + '_' + setup['interface'][0:4] + '_' + tstr1[4:] 
      lfname = tstr1 + '_log' + '.txt'
      log = open(lfname, 'a')
      print >> log, tstr
        
    if save_log == 1:
      print >> log, 'Test Setup: ' 
      print >> log, setup['board'] + ' '
      print >> log, setup['iostandard'] + ' '
      print >> log, setup['idelay_status'] + ' '
      print >> log, setup['iddr_status'] + ' ' 
      print >> log, setup['interface'] + ' ' 
      print >> log, 'FPGA Testbed Version = ' + setup['version']
      print '\n'
  
    u = '~'
    while 1:
      print_cmds()
      u = raw_input('\nEnter Command => ')
      # Quit the test
      if u.upper() == 'Q':
        break
        
      # Print the current Si570 frequency measurement  
      elif u.upper() == 'M':
        show_freq()
        
      # Set a new Si570 frequency    
      elif u.upper() == 'N':
        fnew = raw_input('Enter New Frequency in Hz => ')
        print fnew
        t = create_params(int(fnew))
        print '\nNew Frequency Parameters : ' ,t
        if save_log == 1:
          print >> log, '\nNew Frequency Parameters : ' ,t
        print 'Setting New Output Frequency'
        write_config(t[fparams_config])
        config_current = read_config()
        time.sleep(1) # delay to see frequency change on scope if desired
        show_freq()
    
      elif u.upper() == 'S':
        fnew0 = raw_input('Enter Start Frequency in Hz => ')
        fnew1 = raw_input('Enter End Frequency in Hz => ')
        step = raw_input('Enter Frequency Spacing in Hz => ')
        if save_log == 1:
          print >> log, 'Frequency Sweep From ', fnew0, ' to ', fnew1, ' at ', step, ' Intervals'
        f0 = int(fnew0)
        f1 = int(fnew1)
        f = f0
        while f < (f1 + 1):      
          t = create_params(f)
          print '\nNew Frequency = ', f, 'New Frequency Parameters : ' ,t
          if save_log == 1:
            print >> log,'\nNew Frequency = ', f, 'New Frequency Parameters : ' ,t
          print 'Setting New Output Frequency'
          write_config(t[fparams_config])
          config_current = read_config()
          time.sleep(1) # delay to see frequency change on scope if desired
          show_freq()
          run_test()
          time.sleep(1)               
          read_status()
          v = get_results()
          if v[0] == 1048576:
            print 'Test Passed'
            if save_log == 1:
              print >> log, 'Test Passed'
            x = float(v[1])/100000000.0
            print 'Elapsed Time = ',
            print x,
            print ' seconds'
            if save_log == 1:
              print >> log, 'Elapsed Time = ',
              print >> log, x,
              print >> log, ' seconds'
            y = 4.0*1048576.0*100.0/float(v[1])
            print 'data Rate = ',
            print y,
            print ' MB/s'
            if save_log == 1:
              print >> log, 'data Rate = ',
              print >> log, y,
              print >> log, ' MB/s'
          else:
            print '***Test Failed***'  
            if save_log == 1:
              print >> log, '***Test Failed***'  
          f = f + int(step)
        
      # Run a test at the current frequency    
      elif u.upper() == 'R':
        v = [0,0]
        run_test()
        time.sleep(1)               
        read_status()
        v = get_results()
        if v[0] == 1048576:
          print 'Test Passed'
          if save_log == 1:
            print >> log, 'Test Passed'
          x = float(v[1])/100000000.0
          print 'Elapsed Time = ',
          print x,
          print ' seconds'
          if save_log == 1:
            print >> log, 'Elapsed Time = ',
            print >> log, x,
            print >> log, ' seconds'
          y = 4.0*1048576.0*100.0/float(v[1])
          print 'data Rate = ',
          print y,
          print ' MB/s'
          if save_log == 1:
            print >> log, 'data Rate = ',
            print >> log, y,
            print >> log, ' MB/s'
        else:
          print '***Test Failed***'  
          if save_log == 1:
            print >> log, '***Test Failed***'  
        
      # Run the Data Source continuously in Eye Test Mode   
      elif u.upper() == 'E':
        run_eye()
      
      # Print Control/Status registers    
      elif u.upper() == 'Z':
        read_status()
        
      # Clear registers  
      elif u.upper() == 'C':
        clear_status()
        
      # Force testbed into an Idle state   
      elif u.upper() == 'T':
        terminate_test()
      
      # Set channel iDelays       
      elif u.upper() == 'D':
        channel_idelays = set_delays(channel_idelays)
      
      # Write Reg10  ( not advertised in the Menu as it's experimental )     
      elif u.upper() == 'W':
        write_xadc()
      
      # Display the FPGA core temperature        
      elif u.upper() == 'X':
        get_xadc_temp()
          
      else:
        print '\nNot a valid Command'  
    
    terminate_test
  if save_log == 1:
    log.close    
  ser.close

