-------------------------------------------------------------------------------
--  Copyright (c) 2016 Eclektek LLC. All rights reserved.
--
--
--  Developed by: William Carter
--
--  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and  
--  associated documentation files (the "Software"), to deal with the Software without restriction,  
--  including without limitation the rights to use, copy, modify, merge, publish, sublicense and/or    
--  distribute copies of the Software, and to permit persons to whom the Software is furnished to do so,  
--  subject to the following conditions:
--
--      Redistributions of source code must retain the above copyright notice, this list of conditions 
--      and the following disclaimers.
--      Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
--      and the following disclaimers in the documentation and/or other materials provided with the distribution.
--      Neither the names of William Carter, Eclektek LLC, nor the names of its 
--      contributors may be used to endorse or promote products derived from this Software without 
--      specific prior written permission from Eclektek LLC.
--      This Software is provided free of charge and Eclektek LLC reserves the right to sell copies of the 
--      Software in it original form. Anyone sub-licensing this Software will provide it free of charge.
--
--  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
--  LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
--  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
--  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
--  SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
-------------------------------------------------------------------------------
--  History :                                                                       
--                                                                                                                  
--  Date      By    Comments                                         
--  --------  ----  ------------------------------------------------------------      
--  09-13-16  W.C.  Created                                     
--
-------------------------------------------------------------------------------
--                                                                                                                                                                     
--  Project       : Differential PMOD Challenge                                                           
--  PCB           : Nexys Video                  
--  Entity Name   : Crude_CDR
--  File Name     : Crude_CDR.vhd
--  Design Parent : none                                                          
--                                                                             
-------------------------------------------------------------------------------
--  Dependencies:
--
--  Crude_CDR.vhd is uses the following components:
--  Component      none
--
--  Crude_CDR.vhd is uses the following packages:
--  Package        none
--
-------------------------------------------------------------------------------
--  Description:
--    
--    Crude_CDR is, as the name suggests, a crude data and clock recovery design.
--    This implementation was created specifically for the DataSink_BiPH_L.
--
--    The idea here is to implement a sort of DPLL to derive a reference clock
--    from the sampled data. As we are expecting RZ bi-phase level encoded di-bits
--    we expect a transition will, when it occurs, to be in the middle of a data
--    bit period. By looking for changes in the incoming data we can detect these
--    transitions and derive a clock corresponding to the di-bit period. From this
--    clock reference we can create a bit-period clock. This implementation assumes
--    that each of the data signals have different jitter and delay characteristics, 
--    and that is there is no expectation that all of the transitions occur in the same 
--    fs_clk period for all 4 data signals. It is essential to detect a loss of lock 
--    when incoming data stops changing so that the receiver can stop trying to derive data
--    from the CDR data. For this project we have a wide range of data rates to support  
--    making the "LOS" detection somewhat complicated and sub-optimal. In a normal
--    application a restricted range or number of clock rates would be covered. The
--    means of determining the di-bit clock period is simply to count fs_clk periods
--    during the training preamble phase ( when data is chosen to guarantee a transition
--    every data bit period ). Unfortunately, we can't expect an unknown data bit period
--    to be an integer number of fs_clock periods. For this implementation we don't
--    deal with fractional bit periods ( though we could... ) so the derived data bit
--    period will vary from bit to bit. This represents one simplification that has an
--    adverse effect on performance. Another issue is that in a normal PLL we'd have a
--    reference VCO at a much greater frequency than the recovered clock frequency. Due
--    to limitations on how fast our FPGA can clock logic and meet timing, we face another
--    limitation affecting performance. As the number of samples per bit period gets
--    lower than 6 ( 3 per di-bit ) it will be difficult to locate the position of a
--    consitantly good sample for extracting data. This is true even if we have a perfect
--    channel without overshoot, undershoot, data bit skew, reflections and the like...
--    and we will certainly have most if not all of these. The end result is that this
--    implementation is highly sensitive to data rates that are above fs_clk/6 and especially
--    to data rates with bit periods not equal to an integer number of fs_clk periods. 
--
--    Once a data clock has been derived we need to synchronize our derived clock to 
--    the actual data transitions. This is complicated by not having all of the bits 
--    necessarily transition in the same fs_clk period. This implementation uses the
--    latest transition among the data bits to mark where we derive data. At data clock rates 
--    above fs_clk/6 we don't have a lot of samples to play with. Any overshoot or undershoot 
--    in the signals coming into the FPGA SinkPMOD pins is likely to cause errors. This 
--    implemenation simply uses the latest data transition to reset the derived data clock counter 
--    and in effect synchronize the clock to the center of a bit period. This is another
--    simplification resulting in a performance penalty, especially at high data rates. 
--
--    Phasing for this implementation depends on the the sync word msbs being '1' 
--
-------------------------------------------------------------------------------
--  Known unresolved Issues:
-------------------------------------------------------------------------------
--  Deficiencies
-------------------------------------------------------------------------------
--  Limitations
--    This design by no means represents an optimal high performance CDR design
--
-------------------------------------------------------------------------------
--  ToDo:
-------------------------------------------------------------------------------
--  NOTES:
-------------------------------------------------------------------------------



Library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_arith.all;
  use ieee.std_logic_unsigned.all;

    
entity Crude_CDR is
  port(
        -- GLOBAL SIGNALS
        fs_clk                          : in    std_logic;                     -- sample clock
        areset                          : in    std_logic;                     -- asynchronous reset
                
        data_in                         : in    std_logic_vector(3 downto 0);  -- sampled data
        sync_word_detected              : in    std_logic;                       
        data_out                        : out   std_logic_vector(3 downto 0);  -- recovered data
        data_shft_en                    : out   std_logic;
        los                             : out   std_logic
        
  );
end entity;

architecture rtl of Crude_CDR is

-------------------------------------------------------------------------------
--  Component Declarations
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--  Type Declarations
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--  Constant Declarations
-------------------------------------------------------------------------------

constant  SYNC_PATTERN          : std_logic_vector(31 downto 0) := X"AAAA5569";
constant  PREAMBLE_PATTERN      : std_logic_vector(31 downto 0) := X"AAAAAAAA";
constant  IDLE_PATTERN          : std_logic_vector(31 downto 0) := X"00000000";

-------------------------------------------------------------------------------
--  Signal Declarations
-------------------------------------------------------------------------------

signal  data_pipe_0       : std_logic_vector(3 downto 0);
signal  data_pipe_1       : std_logic_vector(3 downto 0);
signal  data_pipe_2       : std_logic_vector(3 downto 0);
signal  data_pipe_3       : std_logic_vector(3 downto 0);
signal  data_pipe_4       : std_logic_vector(3 downto 0);
signal  bit_center_stb    : std_logic;
signal  bit_center_stb_r  : std_logic;
signal  bit_center_stb_r1 : std_logic;
signal  sample_stb        : std_logic;
signal  phase             : std_logic;
signal  data_ph0          : std_logic_vector(3 downto 0);
signal  data_ph1          : std_logic_vector(3 downto 0);
signal  data_ph           : std_logic_vector(3 downto 0);
signal  data_ld0          : std_logic;
signal  data_ld1          : std_logic;
signal  bclk_per_cnt      : std_logic_vector(11 downto 0);
signal  bclk_timeout_cnt  : std_logic_vector(11 downto 0);
signal  bclk_lost         : std_logic;
signal  bclk_cnt          : std_logic_vector(3 downto 0);
signal  max_blck_per      : std_logic_vector(11 downto 0);
signal  max_blck_halfper  : std_logic_vector(11 downto 0);
signal  locked_clk_cntr   : std_logic_vector(11 downto 0);
signal  locked_clk        : std_logic;
signal  locked_data_clk   : std_logic;
signal  locked_data_clk_r : std_logic;
signal  shift_en          : std_logic;
signal  data_sel          : std_logic;
signal  sm_pipe           : std_logic_vector(5 downto 0);
signal  is_idle           : std_logic;
signal  is_acquiring      : std_logic;
signal  is_locking        : std_logic;

-------------------------------------------------------------------------------
--  Procedure Declarations
-------------------------------------------------------------------------------
  
  
begin


-------------------------------------------------------------------------------
--  Component Instantiations
-------------------------------------------------------------------------------

  
-------------------------------------------------------------------------------
--  Concurrent Assignments
-------------------------------------------------------------------------------

  bit_center_stb  <= '1' when bit_center_stb_r = '0' and bit_center_stb_r1 = '1' else '0';
  bclk_lost       <= '1' when bclk_timeout_cnt > X"1FE" else 
                     '0';
  max_blck_halfper  <= '0' & max_blck_per(11 downto 1);
  locked_clk      <= '0' when locked_clk_cntr < max_blck_halfper else '1';
  data_ld0        <= '1' when locked_data_clk = '0' and locked_data_clk_r = '1' else '0'; 
  data_ld1        <= '1' when locked_data_clk = '1' and locked_data_clk_r = '0' else '0'; 
  shift_en        <= '1' when locked_data_clk = '1' and locked_data_clk_r = '0' else '0'; 
  data_ph         <= data_ph0 when data_sel = '0' else data_ph1;
  is_idle       <= '1' when sm_pipe = "000000" else '0';
  is_acquiring  <= '1' when sm_pipe = "000001" else '0';
  is_locking    <= '1' when sm_pipe = "000011" else '0';
  
-------------------------------------------------------------------------------
--  Process Declarations
-------------------------------------------------------------------------------
   
  -- store a number of copies of the input data to account for delays in processing
  process(fs_clk,areset)
  begin
    if areset = '1' then
      data_pipe_0  <= (others => '1');  
      data_pipe_1  <= (others => '1');  
      data_pipe_2  <= (others => '1');  
      data_pipe_3  <= (others => '1');  
      data_pipe_4  <= (others => '1');  
    elsif (rising_edge(fs_clk)) then
      data_pipe_0  <= data_in;  
      data_pipe_1  <= data_pipe_0;  
      data_pipe_2  <= data_pipe_1;  
      data_pipe_3  <= data_pipe_2;  
      data_pipe_4  <= data_pipe_3;  
    end if;
  end process; 

  -- before a derived reference data bit clock exists start marking data transitions
  -- we are assured that there will be one every preamble data bit period
  process(fs_clk,areset)
  begin
    if areset = '1' then
      bit_center_stb_r    <= '0';  
      bit_center_stb_r1   <= '0';  
    elsif (rising_edge(fs_clk)) then
      bit_center_stb_r1   <= bit_center_stb_r;
      if data_pipe_0 /= data_in then
        bit_center_stb_r    <= '1';
      elsif data_pipe_0 = data_in then   
        bit_center_stb_r    <= '0';
      end if;  
    end if;
  end process; 
  
  -- create an interim phase from the preamble transitions  
  phase_proc : process(fs_clk,areset)
  begin
    if areset = '1' then
      phase <= '0';
    elsif (rising_edge(fs_clk)) then
      if bclk_lost = '1' then 
        phase <= '0';
      elsif bit_center_stb = '1' then            
        phase <= not phase;
      end if;  
    end if;                                                                       
  end process phase_proc; 

  -- measure the di-bit period
  process(fs_clk,areset)
  begin
    if areset = '1' then
      bclk_per_cnt  <= (others => '0');  
    elsif (rising_edge(fs_clk)) then
      if bit_center_stb = '1' then
        bclk_per_cnt  <= (others => '0');  
      else
        bclk_per_cnt <= bclk_per_cnt + 1;
      end if;
    end if;
  end process; 

  -- save the maximum di-bit period to create a derived reference clock
  process(fs_clk,areset)
  begin
    if areset = '1' then
      max_blck_per  <= (others => '0');  
    elsif (rising_edge(fs_clk)) then
      if is_idle = '1' then
        max_blck_per  <= (others => '0');  
      elsif  is_acquiring = '1' and max_blck_per < bclk_per_cnt then 
        max_blck_per <= bclk_per_cnt;
      end if;
    end if;
  end process; 

  -- count preamble di-bits to allow determining phasing before we need
  -- to start looking for the sync word in the data shift register using
  -- our derived reference clock
  process(fs_clk,areset)
  begin
    if areset = '1' then
      bclk_cnt  <= (others => '0');  
    elsif (rising_edge(fs_clk)) then
      if is_idle = '1' then
        bclk_cnt  <= (others => '0');  
      elsif  is_acquiring = '1' and bit_center_stb = '1' and phase = '1' then 
        bclk_cnt <= bclk_cnt + 1;
      elsif  is_locking = '1' and bit_center_stb = '1' and phase = '1' then 
        bclk_cnt <= bclk_cnt + 1;
      end if;
    end if;
  end process; 
  
  -- the derived di-bit clock
  process(fs_clk,areset)
  begin
    if areset = '1' then
      locked_clk_cntr  <= (others => '0');  
    elsif (rising_edge(fs_clk)) then
      if bclk_lost = '1' then 
        locked_clk_cntr  <= (others => '0');
      elsif bit_center_stb = '1' then    
        locked_clk_cntr  <= (others => '0');  
      elsif locked_clk_cntr < max_blck_per then 
        locked_clk_cntr <= locked_clk_cntr + '1';
      else  
        locked_clk_cntr  <= (others => '0');  
      end if;
    end if;
  end process; 

  -- the derived data bit reference clock
  process(fs_clk,areset)
  begin
    if areset = '1' then
      locked_data_clk  <= '0';  
    elsif (rising_edge(fs_clk)) then
      if locked_clk_cntr = max_blck_halfper then
        locked_data_clk  <= not locked_data_clk;
      end if;    
    end if;
  end process; 
  
  -- a delayed copy of the reference clock
  process(fs_clk,areset)
  begin
    if areset = '1' then
      locked_data_clk_r  <= '0';  
    elsif (rising_edge(fs_clk)) then
      locked_data_clk_r  <= locked_data_clk;
    end if;
  end process; 

  -- an LOS detector
  -- always look for an absence of transitions in the data exceeding
  -- the longest possible di-bit bit period
  process(fs_clk,areset)
  begin
    if areset = '1' then
      bclk_timeout_cnt  <= (others => '0');  
    elsif (rising_edge(fs_clk)) then
      if bit_center_stb = '1' then
        bclk_timeout_cnt  <= (others => '0');  
      elsif bclk_timeout_cnt < X"2FF" then 
        bclk_timeout_cnt <= bclk_timeout_cnt + 1;
      end if;
    end if;
  end process; 

  -- save data from the center of the first di-bit
  process(fs_clk,areset)
  begin
    if areset = '1' then
      data_ph0  <= (others => '0');  
    elsif (rising_edge(fs_clk)) then
      if data_ld0 = '1'  then
        data_ph0 <= data_pipe_4;
      end if;
    end if;
  end process; 

  -- save data from the center of the second di-bit
  process(fs_clk,areset)
  begin
    if areset = '1' then
      data_ph1  <= (others => '0');  
    elsif (rising_edge(fs_clk)) then
      if data_ld1 = '1'  then
        data_ph1 <= data_pipe_4;
      end if;
    end if;
  end process; 
 
  -- during preamble we need to get the phasing correct
  -- knowing what to expect helps
  process(fs_clk,areset)
  begin
    if areset = '1' then
      data_sel  <= '0';  
    elsif (rising_edge(fs_clk)) then
      if is_locking = '1' then
        if data_ph0 = X"A"  then
          data_sel <= '0';
        else
          data_sel <= '1';
        end if;
      end if;
    end if;
  end process; 
  
   
  -- simplistic state machine
  --                        
  -- 000000 idle when run_en = '0'
  -- 000001 acquire bit clock
  -- 000011 lock bit clock
  -- 000111 acquire sync word  
  -- 001111 detect sync word  
  -- 011111 data
  -- 111111 detect loss of bit clock
  sm_proc: process(fs_clk,areset)
  begin
    if areset = '1' then
      sm_pipe <= (others => '0');
    elsif (rising_edge(fs_clk)) then
      if bclk_lost = '1' then
        sm_pipe <= (others => '0');
      elsif is_idle = '1' and data_pipe_0 /= data_in  then
        sm_pipe <= sm_pipe(4 downto 0) & '1';
      elsif is_acquiring = '1' and bclk_cnt = X"E" then
        sm_pipe <= sm_pipe(4 downto 0) & '1';
      elsif is_locking = '1' and bclk_cnt > X"E" then
        sm_pipe <= sm_pipe(4 downto 0) & '1';
      elsif sm_pipe = "000111" and sync_word_detected = '1' then
        sm_pipe <= sm_pipe(4 downto 0) & '1';
      elsif sm_pipe = "001111" and sync_word_detected = '1' then
        sm_pipe <= sm_pipe(4 downto 0) & '1';
      end if;
    end if;
  end process sm_proc;

-------------------------------------------------------------------------------
--  Other Output Assignments
-------------------------------------------------------------------------------
 
  data_out        <= data_ph;
  data_shft_en    <= shift_en;
  los             <= bclk_lost; 
  
end rtl;
