-------------------------------------------------------------------------------
--  Copyright (c) 2016 Eclektek LLC. All rights reserved.
--
--
--  Developed by: William Carter
--
--  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and  
--  associated documentation files (the "Software"), to deal with the Software without restriction,  
--  including without limitation the rights to use, copy, modify, merge, publish, sublicense and/or    
--  distribute copies of the Software, and to permit persons to whom the Software is furnished to do so,  
--  subject to the following conditions:
--
--      Redistributions of source code must retain the above copyright notice, this list of conditions 
--      and the following disclaimers.
--      Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
--      and the following disclaimers in the documentation and/or other materials provided with the distribution.
--      Neither the names of William Carter, Eclektek LLC, nor the names of its 
--      contributors may be used to endorse or promote products derived from this Software without 
--      specific prior written permission from Eclektek LLC.
--      This Software is provided free of charge and Eclektek LLC reserves the right to sell copies of the 
--      Software in it original form. Anyone sub-licensing this Software will provide it free of charge.
--
--  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
--  LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
--  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
--  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
--  SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
-------------------------------------------------------------------------------
--  History :                                                                       
--                                                                                                                  
--  Date      By    Comments                                         
--  --------  ----  ------------------------------------------------------------      
--  09-13-16  W.C.  Created                                     
--
-------------------------------------------------------------------------------
--                                                                                                                                                                     
--  Project       : Differential PMOD Challenge                                                           
--  PCB           : Nexys Video                  
--  Entity Name   : DataSink_BiPH_L
--  File Name     : DataSink_BiPH_L.vhd
--  Design Parent : none                                                          
--                                                                             
-------------------------------------------------------------------------------
--  Dependencies:
--
--  DataSink_BiPH_L.vhd is uses the following components:
--  Crude_CDR      Crude_CDR.vhd
--
--  DataSink_BiPH_L.vhd is uses the following packages:
--  Package        none
--
-------------------------------------------------------------------------------
--  Description:
--    
--    DataSink_BiPH_L is the data receiver for a differential 4-bit interace.
--    This entity deserializes a 32-bit data word from a stream of 8 4-bit symbol chunks.
--    As all of the interface pins are used for data an RZ bi-phase level PCM waveform
--    is used to convey symbols.    
--
--    In this implementation the receiver gets sampled versions of the data. 
--    The objectives here are first to derive a phase-locked clock reference from
--    a training preamble. This is performed using a crude clock recovery design.
--    The reference clock period is created at the di-bit symbol period or 2x the data period.
--    Once a data bit period reference clock has been recovered the receiver needs to determing
--    where the center of the di-bit period is to extract the data from the sampled data.  
--    The receiver looks for a sync word to mark the first chunk of the first 32-bit data word. 
--    Di-bit symbols are then decoded to create data bits in 4-bit chunks. In this implementation 
--    the phase0 data is used as data. Obviously, maintaining di-bit phase is really important 
--    to extracting the correct data from the symbol stream.
--
-------------------------------------------------------------------------------
--  Known unresolved Issues:
-------------------------------------------------------------------------------
--  Deficiencies
-------------------------------------------------------------------------------
--  Limitations
--    This design by no means represents the best RZ bi-phase level PCM waveform  
--    receiver design that can be implemented
--
-------------------------------------------------------------------------------
--  ToDo:
-------------------------------------------------------------------------------
--  NOTES:
-------------------------------------------------------------------------------



Library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_arith.all;
  use ieee.std_logic_unsigned.all;
  use ieee.math_real.all;

    
entity DataSink_BiPH_L is
  port(
        -- GLOBAL SIGNALS
        fs_clk                          : in    std_logic;  
        areset                          : in    std_logic; -- asynchronous reset
                
        -- PMOD 4 bits data
        data_in                         : in    std_logic_vector(3 downto 0);        
        
        -- Received data
        data_out                        : out   std_logic_vector(31 downto 0); -- data received
        pdata_out                       : out   std_logic_vector(3 downto 0);  -- control to testbed
        fifo_full                       : in    std_logic;                     -- '0' if FIFO can receive data
        fifo_wren                       : out   std_logic;                     -- fifo write enable
        
        -- debug
        d                               : out   std_logic_vector(31 downto 0); -- data received
        d_valid                         : out   std_logic        
  );
end entity;

architecture rtl of DataSink_BiPH_L is

-------------------------------------------------------------------------------
--  Component Declarations
-------------------------------------------------------------------------------

component Crude_CDR
  port(
        -- GLOBAL SIGNALS
        fs_clk                          : in    std_logic;                     -- sample clock
        areset                          : in    std_logic;                     -- asynchronous reset
                
        data_in                         : in    std_logic_vector(3 downto 0);  -- sampled data
        sync_word_detected              : in    std_logic;                       
        data_out                        : out   std_logic_vector(3 downto 0);  -- recovered data
        data_shft_en                    : out   std_logic;
        los                             : out   std_logic        
  );
end component;


-------------------------------------------------------------------------------
--  Type Declarations
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--  Constant Declarations
-------------------------------------------------------------------------------

constant  NUM_INPUT_BITS        : natural := 4;
constant  NUM_CHUNKS            : natural := integer(ceil(real(32/NUM_INPUT_BITS)));  
constant  SHIFT_REGISTER_WIDTH  : natural := NUM_INPUT_BITS*NUM_CHUNKS;  -- 32 bits data plus 1 bit control
constant  SYNC_PATTERN          : std_logic_vector(31 downto 0) := X"AAAA5569";
constant  PREAMBLE_PATTERN      : std_logic_vector(31 downto 0) := X"AAAAAAAA";
constant  IDLE_PATTERN          : std_logic_vector(31 downto 0) := X"00000000";

-------------------------------------------------------------------------------
--  Signal Declarations
-------------------------------------------------------------------------------

signal  sdata             : std_logic_vector(SHIFT_REGISTER_WIDTH-1 downto 0);
signal  sync_det          : std_logic;
signal  chunk_cnt         : natural;
signal  data_valid        : std_logic;
signal  end_of_data       : std_logic;
signal  data              : std_logic_vector(3 downto 0);
signal  sdata_shift_en    : std_logic;
signal  loss_of_signal    : std_logic;

-------------------------------------------------------------------------------
--  Procedure Declarations
-------------------------------------------------------------------------------
  
  
begin


-------------------------------------------------------------------------------
--  Component Instantiations
-------------------------------------------------------------------------------

  CDR_inst : Crude_CDR
  port map(
        fs_clk              => fs_clk,
        areset              => areset,               
        data_in             => data_in,
        sync_word_detected  => sync_det,
        data_out            => data,
        data_shft_en        => sdata_shift_en,
        los                 => loss_of_signal
  );
  
-------------------------------------------------------------------------------
--  Concurrent Assignments
-------------------------------------------------------------------------------

  sync_det      <= '1' when sdata(SHIFT_REGISTER_WIDTH-1 downto 0) = SYNC_PATTERN else '0';
  
  
  
-------------------------------------------------------------------------------
--  Process Declarations
-------------------------------------------------------------------------------
  
 
  -- de-serialize whatever is at data_in 
  DATA_DESERIALIZER_PROC : process(fs_clk,areset)
  begin
    if areset = '1' then
      sdata <= (others => '0');
    elsif (rising_edge(fs_clk)) then
      if sdata_shift_en = '1' then 
        sdata <= sdata(SHIFT_REGISTER_WIDTH-1-NUM_INPUT_BITS downto 0) & data;
      else
        sdata <= sdata;
      end if;  
    end if;
  end process DATA_DESERIALIZER_PROC; 
  
  chunk_count_proc : process(fs_clk,areset)
  begin
    if areset = '1' then
      chunk_cnt <= 0;
    elsif (rising_edge(fs_clk)) then
      if sdata_shift_en = '1' then
        if sync_det = '1' then            
          chunk_cnt <= 0;
        elsif chunk_cnt < NUM_CHUNKS-1 then            
          chunk_cnt <= chunk_cnt + 1;
        else  
          chunk_cnt <= 0;
        end if;
      end if;    
    end if;                                                                       
  end process chunk_count_proc;                                                    
  
  process(fs_clk,areset)
  begin
    if areset = '1' then
      data_valid <= '0';
    elsif (rising_edge(fs_clk)) then
      if sdata_shift_en = '1' then
        if sync_det = '1' then            
          data_valid <= '1';
        else  
          data_valid <= data_valid;
        end if;  
      end if;    
    end if;                                                                       
  end process;
  

-------------------------------------------------------------------------------
--  Other Output Assignments
-------------------------------------------------------------------------------
  
  data_out    <= sdata(SHIFT_REGISTER_WIDTH-1 downto 0);
  pdata_out   <= "0000";
  fifo_wren   <= '1' when chunk_cnt = NUM_CHUNKS-1 and data_valid = '1' and sdata_shift_en = '1' else '0'; 

  d           <= sdata(SHIFT_REGISTER_WIDTH-1 downto 0);
  d_valid     <= '1' when chunk_cnt = NUM_CHUNKS-1 else '0'; 
  
end rtl;
