-------------------------------------------------------------------------------
--  Copyright (c) 2016 Eclektek LLC. All rights reserved.
--
--
--  Developed by: William Carter
--
--  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and  
--  associated documentation files (the "Software"), to deal with the Software without restriction,  
--  including without limitation the rights to use, copy, modify, merge, publish, sublicense and/or    
--  distribute copies of the Software, and to permit persons to whom the Software is furnished to do so,  
--  subject to the following conditions:
--
--      Redistributions of source code must retain the above copyright notice, this list of conditions 
--      and the following disclaimers.
--      Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
--      and the following disclaimers in the documentation and/or other materials provided with the distribution.
--      Neither the names of William Carter, Eclektek LLC, nor the names of its 
--      contributors may be used to endorse or promote products derived from this Software without 
--      specific prior written permission from Eclektek LLC.
--      This Software is provided free of charge and Eclektek LLC reserves the right to sell copies of the 
--      Software in it original form. Anyone sub-licensing this Software will provide it free of charge.
--
--  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
--  LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
--  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
--  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
--  SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
-------------------------------------------------------------------------------
--  History :                                                                       
--                                                                                                                  
--  Date      By    Comments                                         
--  --------  ----  ------------------------------------------------------------      
--  09-05-16  W.C.  Created                                     
--
-------------------------------------------------------------------------------
--  Project       : Differential PMOD Challenge                                                           
--  PCB           : Nexys Video                  
--  Entity Name   : DataSink_SDR
--  File Name     : DataSink_SDR.vhd
--  Design Parent : none                                                          
--                                                                             
-------------------------------------------------------------------------------
--  Dependencies:
--
--  DataSink_SDR.vhd is uses the following components:
--  Component      none
--
--  DataSink_SDR.vhd is uses the following packages:
--  Package        none
--
-------------------------------------------------------------------------------
--  Description:
--
--  DataSink_SDR.vhd
--    DataSink_SDR is the data receiver for the simplest differential 3-bit
--    plus clock reference interace. This entity deserializes a 32-bit data word 
--    from a stream of 11 3-bit chunks. The 33rd bit is used to convey one bit of 
--    control. One 3-bit chunk of data is received each reference clock cycle.
--
--    In this implementation the receiver gets sampled versions of the data and 
--    clock reference. The objectives here are to determine where to sample the
--    data, where the 32-bit word boundaries are in the data stream, when to know 
--    when valid data is being received, and when to stop writing data to the testbed. 
--    In this implementation the receiver is always looking at the data signals
--    waiting for transitions that indicate the start of a message. It does not 
--    currently use the training preamble and waits until the sync symbol word is 
--    received before writing to the fifo. The last bit of the last chunk is the
--    control bit letting the receiver know when to stop writing data to the fifo. 
--    
-------------------------------------------------------------------------------
--  Known unresolved Issues:
-------------------------------------------------------------------------------
--  Deficiencies
-------------------------------------------------------------------------------
--  Limitations
--    This design by no means represents the best single data rate design that can
--    be implemented
--
-------------------------------------------------------------------------------
--  ToDo:
-------------------------------------------------------------------------------
--  NOTES:
-------------------------------------------------------------------------------



Library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_arith.all;
  use ieee.std_logic_unsigned.all;

    
entity DataSink_SDR is
  port(
        -- GLOBAL SIGNALS
        fs_clk                          : in    std_logic;  
        areset                          : in    std_logic; -- asynchronous reset
                
        -- PMOD 3 bits data plus clock
        data_in                         : in    std_logic_vector(3 downto 0);        
        
        -- Received data
        data_out                        : out   std_logic_vector(31 downto 0); -- data received
        pdata_out                       : out   std_logic_vector(3 downto 0);  -- control to testbed
        fifo_full                       : in    std_logic;                     -- '0' if FIFO can receive data
        fifo_wren                       : out   std_logic;                     -- fifo write enable
        
        -- debug
        d                               : out   std_logic_vector(31 downto 0); -- data received
        d_valid                         : out   std_logic        
  );
end entity;

architecture rtl of DataSink_SDR is

-------------------------------------------------------------------------------
--  Component Declarations
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--  Type Declarations
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--  Constant Declarations
-------------------------------------------------------------------------------

constant  NUM_INPUT_BITS        : natural := 3;
constant  NUM_CHUNKS            : natural := 11;  -- 32 bits data plus 1 bit control
constant  SHIFT_REGISTER_WIDTH  : natural := NUM_INPUT_BITS*NUM_CHUNKS;  -- 32 bits data plus 1 bit control
constant  SYNC_PATTERN          : std_logic_vector(31 downto 0) := X"AAAA5569";
constant  PREAMBLE_PATTERN      : std_logic_vector(31 downto 0) := X"AAAAAAAA";
constant  IDLE_PATTERN          : std_logic_vector(31 downto 0) := X"00000000";

-------------------------------------------------------------------------------
--  Signal Declarations
-------------------------------------------------------------------------------

signal  sdata             : std_logic_vector(SHIFT_REGISTER_WIDTH-1 downto 0);
signal  sync_det          : std_logic;
signal  chunk_cnt         : natural;
signal  data_valid        : std_logic;
signal  end_of_data       : std_logic;
signal  rising_edge_stb   : std_logic;
signal  falling_edge_stb  : std_logic;
signal  data_clk          : std_logic;
signal  data_clk_r        : std_logic;
signal  data_clk_r1       : std_logic;

-------------------------------------------------------------------------------
--  Procedure Declarations
-------------------------------------------------------------------------------
  
  
begin


-------------------------------------------------------------------------------
--  Component Instantiations
-------------------------------------------------------------------------------

  
-------------------------------------------------------------------------------
--  Concurrent Assignments
-------------------------------------------------------------------------------

  sync_det    <= '1' when sdata(SHIFT_REGISTER_WIDTH-1 downto 1) = SYNC_PATTERN else '0';
  end_of_data <= '1' when sdata(0) = '0' and chunk_cnt = NUM_CHUNKS-1 and data_valid = '1' else '0';
  data_clk    <= data_in(3);
  rising_edge_stb   <= '1' when data_clk_r = '1' and data_clk_r1 = '0' else '0';
  falling_edge_stb  <= '1' when data_clk_r = '0' and data_clk_r1 = '1' else '0';
  

-------------------------------------------------------------------------------
--  Process Declarations
-------------------------------------------------------------------------------

  process(fs_clk,areset)
  begin
    if areset = '1' then
      data_clk_r  <= '0';  
      data_clk_r1 <= '0'; 
    elsif (rising_edge(fs_clk)) then
      data_clk_r  <= data_clk;  
      data_clk_r1 <= data_clk_r;  
    end if;
  end process; 


  -- de-serialize whatever is at data_in 
  DATA_DESERIALIZER_PROC : process(fs_clk,areset)
  begin
    if areset = '1' then
      sdata <= (others => '0');
    elsif (rising_edge(fs_clk)) then
      if falling_edge_stb = '1' then
        sdata <= sdata(SHIFT_REGISTER_WIDTH-1-NUM_INPUT_BITS downto 0) & data_in(2 downto 0);
      else
        sdata <= sdata;
      end if;  
    end if;
  end process DATA_DESERIALIZER_PROC; 
  
  chunk_count_proc : process(fs_clk,areset)
  begin
    if areset = '1' then
      chunk_cnt <= 0;
    elsif (rising_edge(fs_clk)) then
      if falling_edge_stb = '1' then
        if sync_det = '1' then            
          chunk_cnt <= 0;
        elsif chunk_cnt < NUM_CHUNKS-1 and falling_edge_stb = '1' then            
          chunk_cnt <= chunk_cnt + 1;
        else  
          chunk_cnt <= 0;
        end if;
      end if;    
    end if;                                                                       
  end process chunk_count_proc;                                                    
  
  process(fs_clk,areset)
  begin
    if areset = '1' then
      data_valid <= '0';
    elsif (rising_edge(fs_clk)) then
      if falling_edge_stb = '1' then
        if sync_det = '1' then            
          data_valid <= '1';
        else  
          data_valid <= data_valid;
        end if;  
      end if;    
    end if;                                                                       
  end process;

-------------------------------------------------------------------------------
--  Other Output Assignments
-------------------------------------------------------------------------------
  
  data_out    <= sdata(SHIFT_REGISTER_WIDTH-1 downto 1);
  pdata_out   <= "00" & data_valid & end_of_data;
  fifo_wren   <= '1' when chunk_cnt = NUM_CHUNKS-1 and data_valid = '1' and end_of_data = '0' and falling_edge_stb = '1' else '0'; 

  d           <= sdata(SHIFT_REGISTER_WIDTH-1 downto 1);
  d_valid     <= '1' when chunk_cnt = NUM_CHUNKS-1 else '0'; 
  
end rtl;
