-------------------------------------------------------------------------------
--  Copyright (c) 2016 Eclektek LLC. All rights reserved.
--
--
--  Developed by: William Carter
--
--  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and  
--  associated documentation files (the "Software"), to deal with the Software without restriction,  
--  including without limitation the rights to use, copy, modify, merge, publish, sublicense and/or    
--  distribute copies of the Software, and to permit persons to whom the Software is furnished to do so,  
--  subject to the following conditions:
--
--      Redistributions of source code must retain the above copyright notice, this list of conditions 
--      and the following disclaimers.
--      Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
--      and the following disclaimers in the documentation and/or other materials provided with the distribution.
--      Neither the names of William Carter, Eclektek LLC, nor the names of its 
--      contributors may be used to endorse or promote products derived from this Software without 
--      specific prior written permission from Eclektek LLC.
--      This Software is provided free of charge and Eclektek LLC reserves the right to sell copies of the 
--      Software in it original form. Anyone sub-licensing this Software will provide it free of charge.
--
--  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
--  LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
--  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
--  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
--  SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
-------------------------------------------------------------------------------
--  History :                                                                       
--                                                                                                                  
--  Date      By    Comments                                         
--  --------  ----  ------------------------------------------------------------      
--  09-13-16  W.C.  Created                                     
--
-------------------------------------------------------------------------------
--  Project       : Differential PMOD Challenge                                                           
--  PCB           : Nexys Video                  
--  Entity Name   : DataSource_BiPH_L
--  File Name     : DataSource_BiPH_L.vhd
--  Design Parent : none                                                          
--                                                                             
-------------------------------------------------------------------------------
--  Dependencies:
--
--  DataSource_BiPH_L.vhd is uses the following components:
--  Component      none
--
--  DataSource_BiPH_L.vhd is uses the following packages:
--  Package        none
--
-------------------------------------------------------------------------------
--  Description:
--
--  DataSource_BiPH_L implements a 4-bit data transmitter using a self-clocking
--  RZ bi-phase level PCM waveform. This can be thought of as a phase encoded modulator
--  where there is a pulse in either the first half or the second half of a bit
--  period depending on whether a one or zero is being transmitted. The transition
--  in the middle of each bit period implies a reference clock. Depending on how the
--  data changes this transition may not happen every bit period. This implementation
--  sends a training preamble to allow the the receiver a chance to recover the implied
--  reference clock, a sync symbol to mark the beginning chunk of the first valid data
--  word and data. It serializes each 32-bit message word into 8 4-bit chunks. 
--
--  For this implementation we know that the testbed will fill the fifo faster than
--  the transmitter needs it so the fifo_empty signal serves as a transmit enable
--  for the transmitter. The testbed sends words after the test run as ended successfully
--  so that the transmitter and receiver have a chance to return to an idle state.
--   
--  The fifo that allows the testbed to communicate to the transmitter has 4 parity
--  bits that are available for communicating control. This implementation does not
--  send these along to the receiver though it could by adding an additional chunk to
--  the data stream at a loss of efficiency.
--  
--  In this implementation the phase modulation is viewed as a di-bit symbol. Di-bit symbols
--  are transmitted at the tclk rate making the data bit rate 1/2 the tclk rate.   
--    
-------------------------------------------------------------------------------
--  Known unresolved Issues:
-------------------------------------------------------------------------------
--  Deficiencies
-------------------------------------------------------------------------------
--  Limitations
--    This design by no means represents the best RZ bi-phase level PCM waveform  
--    transmitter design that can be implemented
--
-------------------------------------------------------------------------------
--  ToDo:
-------------------------------------------------------------------------------
--  NOTES:
-------------------------------------------------------------------------------



Library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_arith.all;
  use ieee.std_logic_unsigned.all;
  use ieee.math_real.all;
    
entity DataSource_BiPH_L is
  port(
        -- GLOBAL SIGNALS
        tclk                            : in    std_logic; -- variable test clock
        areset                          : in    std_logic; -- asynchronous reset
        
        -- interface  
        data_in                         : in    std_logic_vector(31 downto 0); -- data to be transmitted
        pdata_in                        : in    std_logic_vector(3 downto 0);  -- control from testbed
        fifo_empty                      : in    std_logic;                     -- '1' indiates data available
        fifo_rden                       : out   std_logic;                     -- fifo read enable        
                
        -- PMOD 4 bits data
        data_out                        : out   std_logic_vector(3 downto 0)        
  );
end entity;

architecture rtl of DataSource_BiPH_L is

-------------------------------------------------------------------------------
--  Component Declarations
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--  Type Declarations
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--  Constant Declarations
-------------------------------------------------------------------------------

constant  NUM_OUTPUT_BITS       : natural := 4;
constant  NUM_CHUNKS            : natural := integer(ceil(real(32/NUM_OUTPUT_BITS)));  
constant  SHIFT_REGISTER_WIDTH  : natural := NUM_OUTPUT_BITS*NUM_CHUNKS;  
constant  SYNC_PATTERN          : std_logic_vector(31 downto 0) := X"AAAA5569";
constant  PREAMBLE_PATTERN      : std_logic_vector(31 downto 0) := X"AAAAAAAA";
constant  IDLE_PATTERN          : std_logic_vector(31 downto 0) := X"00000000";

-------------------------------------------------------------------------------
--  Signal Declarations
-------------------------------------------------------------------------------

signal  sdata             : std_logic_vector(SHIFT_REGISTER_WIDTH-1 downto 0);
signal  data              : std_logic_vector(31 downto 0);
signal  sm_pipe           : std_logic_vector(5 downto 0);
signal  chunk_cnt         : natural; 
signal  data_ld           : std_logic;
signal  run_en            : std_logic;
signal  phase             : std_logic;
signal  dout_r            : std_logic_vector(3 downto 0);

-------------------------------------------------------------------------------
--  Procedure Declarations
-------------------------------------------------------------------------------
  
  
begin


-------------------------------------------------------------------------------
--  Component Instantiations
-------------------------------------------------------------------------------

  
-------------------------------------------------------------------------------
--  Concurrent Assignments
-------------------------------------------------------------------------------

  data    <= IDLE_PATTERN     when sm_pipe = "000000" else
             SYNC_PATTERN     when sm_pipe = "001111" else 
             data_in          when sm_pipe = "011111" else 
             data_in          when sm_pipe = "111111" else 
             PREAMBLE_PATTERN;
             
               
  data_ld <= '1' when chunk_cnt = 0 and run_en = '1' and phase = '1' else '0'; 
            
             
-------------------------------------------------------------------------------
--  Process Declarations
-------------------------------------------------------------------------------

  -- serialize whatever is at data_in 
  DATA_SERIALIZER_PROC : process(tclk,areset)
  begin
    if areset = '1' then
      sdata <= (others => '0');
    elsif (rising_edge(tclk)) then
      if data_ld = '1' then  
        sdata <= data;
      elsif phase = '1' then    
        sdata <= sdata(SHIFT_REGISTER_WIDTH-NUM_OUTPUT_BITS-1 downto 0) & X"0";
      end if;  
    end if;
  end process DATA_SERIALIZER_PROC; 
  
  
  
  -- simplistic state machine
  --                        
  -- 000000 idle when run_en = '0'
  -- 000001 preamble
  -- 000011 preamble 
  -- 000111 preamble
  -- 001111 preamble
  -- 011111 sync  
  -- 111111 data
  sm_proc: process(tclk,areset)
  begin
    if areset = '1' then
      sm_pipe <= (others => '0');
    elsif (rising_edge(tclk)) then
      if data_ld = '1' then
        sm_pipe <= sm_pipe(4 downto 0) & '1';
      elsif run_en = '0' then 
        sm_pipe <= (others => '0');
      end if;
    end if;
  end process sm_proc;

  --- count from 0 to 7 to   
  chunk_count_proc : process(tclk,areset)
  begin
    if areset = '1' then
      chunk_cnt <= 0;
    elsif (rising_edge(tclk)) then
      if run_en = '0' then            
        chunk_cnt <= 0;
      elsif chunk_cnt < NUM_CHUNKS-1 and phase = '1' then            
        chunk_cnt <= chunk_cnt + 1;
      elsif chunk_cnt < NUM_CHUNKS-1 and phase = '0' then            
        chunk_cnt <= chunk_cnt;
      elsif chunk_cnt = NUM_CHUNKS-1 and phase = '1' then 
        chunk_cnt <= 0;
      end if;  
    end if;                                                                       
  end process chunk_count_proc;                                                    

  run_proc : process(tclk,areset)
  begin
    if areset = '1' then
      run_en <= '0';
    elsif (rising_edge(tclk)) then
      if fifo_empty = '0' then            
        run_en <= '1';
      elsif fifo_empty = '1' and chunk_cnt = NUM_CHUNKS-1 then            
        run_en <= '0';
      else  
        run_en <= run_en;
      end if;  
    end if;                                                                       
  end process run_proc;                                                    
  
  
  phase_proc : process(tclk,areset)
  begin
    if areset = '1' then
      phase <= '1';
    elsif (rising_edge(tclk)) then
      if run_en = '0' then            
        phase <= '1';
      else  
        phase <= not phase;
      end if;  
    end if;                                                                       
  end process phase_proc; 
  
  
  process(tclk,areset)
  begin
    if areset = '1' then
      dout_r <= (others => '1');
    elsif (rising_edge(tclk)) then
      dout_r <= sdata(SHIFT_REGISTER_WIDTH-1 downto SHIFT_REGISTER_WIDTH-NUM_OUTPUT_BITS) xor (phase & phase & phase & phase);
    end if;                                                                       
  end process phase_proc;  
  
-------------------------------------------------------------------------------
--  Other Output Assignments
-------------------------------------------------------------------------------
  
  data_out(3 downto 0)    <= dout_r;
  
  
  fifo_rden   <= '1' when chunk_cnt = NUM_CHUNKS-1 and sm_pipe = "011111" and phase = '1' else -- fifo has 1 clock latency
                 '1' when chunk_cnt = NUM_CHUNKS-1 and sm_pipe = "111111" and phase = '1' else 
                 '0';  

end rtl;
