-------------------------------------------------------------------------------
--  Copyright (c) 2016 Eclektek LLC. All rights reserved.
--
--
--  Developed by: William Carter
--
--  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and  
--  associated documentation files (the "Software"), to deal with the Software without restriction,  
--  including without limitation the rights to use, copy, modify, merge, publish, sublicense and/or    
--  distribute copies of the Software, and to permit persons to whom the Software is furnished to do so,  
--  subject to the following conditions:
--
--      Redistributions of source code must retain the above copyright notice, this list of conditions 
--      and the following disclaimers.
--      Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
--      and the following disclaimers in the documentation and/or other materials provided with the distribution.
--      Neither the names of William Carter, Eclektek LLC, nor the names of its 
--      contributors may be used to endorse or promote products derived from this Software without 
--      specific prior written permission from Eclektek LLC.
--      This Software is provided free of charge and Eclektek LLC reserves the right to sell copies of the 
--      Software in it original form. Anyone sub-licensing this Software will provide it free of charge.
--
--  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
--  LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
--  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
--  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
--  SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
-------------------------------------------------------------------------------
--  History :                                                                       
--                                                                                                                  
--  Date      By    Comments                                         
--  --------  ----  ------------------------------------------------------------
--  09/05/16  W.C.  Created                                     
--
-------------------------------------------------------------------------------
--  Project       : Differential PMOD Challenge                                                          
--  PCB           : any
--  Entity Name   : Testbed
--  File Name     : Testbed.vhd
--  Design Parent : none                                                          
--                                                                             
-------------------------------------------------------------------------------
--  Dependencies:
--
--  Testbed.vhd is uses the following components:
--    GENESYS2_Si570_INF      GENESYS2_Si570_INF .nco
--    NexysVideo_Si570_INF    NexysVideo_Si570_INF .nco
--
--
--  Testbed.vhd is uses the following packages:
--    Package        none
--
-------------------------------------------------------------------------------
--  Description:
--
--    Testbed provides resources needed to conduct a test run and a consistant interface
--    to the Data Source and Data Sink DUTs:
--      Pseudo-random 32-bit word generators ( Linear feedback shift registers )
--      Si570_INF contains a UART for software control of a test run and access to the test
--        results. It also contains the I2C interface for programming the Si570 device on the
--        FMC connector. The UART is embedded in an ascii debug component that has parameterized
--        number of registers, registers widths, and UART data rate. For the Si570_INF there are
--        16 48-bit registers 4 of which are used by the I2C interface.
--        12 local 48-bit registers are implemented for test control and status
--      FIFO interfaces to pass data between the Testbed and Data Source and Testbed and Data Sink
--      A Timer to measure elapsed time for a test run
--      A state machine to enforce consistant test run behavior
--      A Data checker to detect errors between words sent by the Data Source and received by the 
--        Data Sink
--
--    Testbed instantiates the XADC to monitor the FPGA Temperature sensor. xadc_fan_ctrl is an
--    output that cn abe connected directly to the board FAN_PWN output and or a LED.
-- 
-------------------------------------------------------------------------------
--  Known unresolved Issues:
-------------------------------------------------------------------------------
--  Limitations
-------------------------------------------------------------------------------
--  ToDo:
--    - could be restructured into multiple files for readability...
--
-------------------------------------------------------------------------------
--  NOTES:
-------------------------------------------------------------------------------



Library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_arith.all;
  use ieee.std_logic_unsigned.all;

library UNISIM;
  use UNISIM.vcomponents.all;
  
entity Testbed is
  generic(
      IS_GENESYS2_BOARD        : natural range 1 downto 0 := 1);                -- 1 when targetting the Genesys 2 board
  port(
        -- Testbed interface
        tb_clock_100                    : in    std_logic;                      -- 100 MHz testbed clock
        dut_clk                         : in    std_logic;                      -- programmable DUT Data Source clock
        sample_clk                      : in    std_logic;                      -- Data Sink sample clock unrelated to dut_clk
        areset                          : in    std_logic;                      -- asynchronous reset
        sreset_out                      : out   std_logic;                      -- synchronous reset from UART control 
                                                                                        
        sim_run_words_m1                : in    std_logic_vector(23 downto 0);  -- number of 32-bit words minus 1 in a test run 
        run_words_m1                    : in    std_logic_vector(23 downto 0);  -- number of 32-bit words minus 1 in a test run 
        sim_source_lfsr_seed            : in    std_logic_vector(31 downto 0);  -- anything that ISN'T X"00000000"
        source_lfsr_seed                : in    std_logic_vector(31 downto 0);  -- anything that ISN'T X"00000000"
        test_id                         : in    std_logic_vector(47 downto 0);  -- a number identifying test options  
        
        -- Board User Switch inputs
        simulation_mode_flag            : in    std_logic;                      -- '1' sets simulation mode, '0' sets hardware mode
                
        -- Board User Button inputs
        sim_start_stb                   : in    std_logic;                      -- asserted high to start a simulation test run
        
        
        -- LED suitable indicators
        test_active                     : out   std_logic;                      -- '1' when data is being transfered during a test run
        test_failed                     : out   std_logic;                      -- '1' when sink data /= source data. latched until new test run
        test_ready                      : out   std_logic;                      -- '1' when the test state machine is in the IDLE state 
        tb_clock_100_heartbeat          : out   std_logic;                      -- 1 Hz signal indicating that the clk100 is there 
        dut_clk_heartbeat               : out   std_logic;                      -- low hz signal representing dut_clk/67108864                                   
        xadc_fan_ctrl                   : out   std_logic;                      -- fan control based on temperature sensor
        
        -- PMOD suitable monitor signals                
        si570_div16_monitor             : out   std_logic;                      -- dut_clk/16 for measurement with external equipment
        
        -- I2C PHYSICAL INTERFACE to operate the FMC Si570 programmable clock
        scl                             : out   std_logic;                      -- Si570 I2C clock
        sda_in                          : in    std_logic;                      -- Si570 I2C data in
        sda_out                         : out   std_logic;                      -- Si570 I2C data out to tri-state buffer
        sda_oe                          : out   std_logic;                      -- Si570 I2C data tri-state control
        
        -- UART
        uart_rx_out                     : out   std_logic;                      -- RS-232 data to USB UART 
        uart_tx_in                      : in    std_logic;                      -- RS-232 data from USB UART
        
        -- IDELAY controls
        sink_idelay_ch0                 : out   std_logic_vector(4 downto 0);   -- IDELAY delay control for D0
        sink_idelay_ch1                 : out   std_logic_vector(4 downto 0);   -- IDELAY delay control for D1
        sink_idelay_ch2                 : out   std_logic_vector(4 downto 0);   -- IDELAY delay control for D2
        sink_idelay_ch3                 : out   std_logic_vector(4 downto 0);   -- IDELAY delay control for D3/CLK
                        
        -- Data Source DUT Interface                                                
        source_clk                      : out   std_logic;                      -- source PMOD data bit clock
        source_areset                   : out   std_logic;                      -- source asynchronous reset input
        source_fifo_data                : out   std_logic_vector(31 downto 0);  -- data to be transmitted
        source_fifo_pdata               : out   std_logic_vector(3 downto 0);   -- source control from testbed
        source_fifo_empty               : out   std_logic;                      -- '1' indiates data available
        source_fifo_rden                : in    std_logic;                      -- fifo read enable        
        source_data_out                 : in    std_logic_vector(3 downto 0);   -- to output PMOD    
        
        -- Data Sink DUT  Interface
        sink_areset                     : out   std_logic;                      -- asynchronous reset
        sink_sample_clk                 : out   std_logic;                      -- sample clock for JC signals
        sink_fifo_full                  : out   std_logic;                      -- '0' if FIFO can receive data
        sink_fifo_data                  : in    std_logic_vector(31 downto 0);  -- data received
        sink_fifo_pdata                 : in    std_logic_vector(3 downto 0);   -- sink control to testbed
        sink_fifo_wren                  : in    std_logic                       -- fifo write enable        
  );
end entity;

architecture rtl of Testbed is

-------------------------------------------------------------------------------
--  Component Declarations
-------------------------------------------------------------------------------


  -- IP Provided as Si570.ngc
  component NexysVideo_Si570_INF
    port(
          -- GLOBAL SIGNALS
          clk_100                         : in    std_logic; 
          areset                          : in    std_logic;  
          smreset                         : in    std_logic;
          
          -- General I/O resources
          clk100_heartbeat_led            : out   std_logic;
          clk_si570_heartbeat_led         : out   std_logic;
          si570_div16_mon                 : out   std_logic;
            
          -- UART Debug interface
          reg_address                     : out   std_logic_vector(3 downto 0);
          reg_data_out                    : out   std_logic_vector(47 downto 0);
          reg_write_stb                   : out   std_logic;
          reg_data_in                     : in    std_logic_vector(47 downto 0);
          reg_read_stb                    : out   std_logic;
          busy                            : out   std_logic;
          
          -- I2C PHYSICAL INTERFACE
          scl                             : out   std_logic; 
          sda_in                          : in    std_logic; 
          sda_out                         : out   std_logic; 
          sda_oe                          : out   std_logic; 
          
          -- UART
          uart_rx                         : out   std_logic;
          uart_tx                         : in    std_logic;
          
          -- FMC Si570 clock from IBUFDS
          vclk                            : in    std_logic
    );
  end component;
  
  component Genesys2_Si570_INF
    port(
          -- GLOBAL SIGNALS
          clk_100                         : in    std_logic; 
          areset                          : in    std_logic;  
          smreset                         : in    std_logic;
          
          -- General I/O resources
          clk100_heartbeat_led            : out   std_logic;
          clk_si570_heartbeat_led         : out   std_logic;
          si570_div16_mon                 : out   std_logic;
            
          -- UART Debug interface
          reg_address                     : out   std_logic_vector(3 downto 0);
          reg_data_out                    : out   std_logic_vector(47 downto 0);
          reg_write_stb                   : out   std_logic;
          reg_data_in                     : in    std_logic_vector(47 downto 0);
          reg_read_stb                    : out   std_logic;
          busy                            : out   std_logic;
          
          -- I2C PHYSICAL INTERFACE
          scl                             : out   std_logic; 
          sda_in                          : in    std_logic; 
          sda_out                         : out   std_logic; 
          sda_oe                          : out   std_logic; 
          
          -- UART
          uart_rx                         : out   std_logic;
          uart_tx                         : in    std_logic;
          
          -- FMC Si570 clock from IBUFDS
          vclk                            : in    std_logic
    );
  end component;
     

-------------------------------------------------------------------------------
--  Type Declarations
-------------------------------------------------------------------------------

type sm_states is (IDLE, FIFO_RST, START, RUN, EYE, OHCRAP, ENDRUN, ENDRUN1);

-------------------------------------------------------------------------------
--  Constant Declarations
-------------------------------------------------------------------------------

constant XADC_38_C_FAN_ON_TEMP           : std_logic_vector(15 downto 0) := X"9E0B"; -- 38 degress C
constant XADC_40_C_FAN_ON_TEMP           : std_logic_vector(15 downto 0) := X"9F10"; -- 40 degress C
constant XADC_35_C_FAN_OFF_TEMP          : std_logic_vector(15 downto 0) := X"9C85"; -- 35 degress C

-------------------------------------------------------------------------------
--  Signal Declarations
-------------------------------------------------------------------------------

signal  current_state, next_state        : sm_states; 

signal sreset                     : std_logic;
signal ireg_address               : std_logic_vector(3 downto 0);
signal ireg_data_out              : std_logic_vector(47 downto 0);
signal ireg_write_stb             : std_logic;
signal ireg_data_in               : std_logic_vector(47 downto 0);
signal ireg_read_stb              : std_logic;
signal ibusy                      : std_logic;
signal dbg_reg0_r                 : std_logic_vector(47 downto 0);
signal dbg_reg1_r                 : std_logic_vector(47 downto 0);
signal dbg_reg2_r                 : std_logic_vector(47 downto 0);
signal dbg_reg3_r                 : std_logic_vector(47 downto 0);
signal dbg_reg4_r                 : std_logic_vector(47 downto 0);
signal dbg_reg5_r                 : std_logic_vector(47 downto 0);
signal dbg_reg6_r                 : std_logic_vector(47 downto 0);
signal dbg_reg7_r                 : std_logic_vector(47 downto 0);
signal dbg_reg8_r                 : std_logic_vector(47 downto 0);
signal dbg_reg9_r                 : std_logic_vector(47 downto 0);
signal dbg_reg10_r                : std_logic_vector(47 downto 0);
signal dbg_reg11_r                : std_logic_vector(47 downto 0);
signal reg2_wr_stb                : std_logic;
signal reg2_wr_data               : std_logic_vector(47 downto 0);
signal reg3_wr_stb                : std_logic;
signal reg3_wr_data               : std_logic_vector(47 downto 0);
signal reg4_wr_stb                : std_logic;
signal reg4_wr_data               : std_logic_vector(47 downto 0);
signal reg5_wr_stb                : std_logic;
signal reg5_wr_data               : std_logic_vector(47 downto 0);
signal reg11_wr_stb               : std_logic;
signal reg11_wr_data              : std_logic_vector(47 downto 0);
signal sda_in_i                   : std_logic;
signal sda_out_i                  : std_logic;
signal scl_i                      : std_logic;
signal sda_oe_i                   : std_logic;
signal iclk100_heartbeat_led      : std_logic;
signal iclk_si570_heartbeat_led   : std_logic;
signal source_fifo_din            : std_logic_vector(31 downto 0);   
signal source_fifo_pdin           : std_logic_vector(3 downto 0);   
signal source_fifo_rd             : std_logic;
signal source_fifo_wr             : std_logic;
signal source_fifo_empty_flag     : std_logic;
signal source_fifo_write_count    : std_logic_vector(11 downto 0);
signal source_fifo_reset          : std_logic;
signal source_fifo_empty_tb_clock_100   : std_logic;
signal source_fifo_empty_tb_clock_100_r : std_logic;
signal sink_fifo_clk              : std_logic;
signal sink_fifo_dout             : std_logic_vector(31 downto 0);   
signal sink_fifo_pdout            : std_logic_vector(3 downto 0);   
signal sink_fifo_rd               : std_logic;
signal sink_fifo_empty            : std_logic;
signal sink_fifo_full_flag        : std_logic;
signal sink_fifo_reset            : std_logic;
signal source_lfsr_ld             : std_logic;
signal source_lfsr_en             : std_logic;
signal source_lfsr_data           : std_logic_vector(31 downto 0);
signal sink_lfsr_ld               : std_logic;
signal sink_lfsr_en               : std_logic;
signal sink_lfsr_en_r1            : std_logic;
signal sink_lfsr_en_r2            : std_logic;
signal sink_lfsr_en_r3            : std_logic;
signal sink_lfsr_en_r4            : std_logic;
signal sink_lfsr_data             : std_logic_vector(31 downto 0);
signal lfsr_seed                  : std_logic_vector(31 downto 0);
signal start_test                 : std_logic;                                             
signal error_det                  : std_logic;
signal test_complete              : std_logic;
signal run_enable                 : std_logic;
signal data_counter               : std_logic_vector(23 downto 0); 
signal dcounter_ld                : std_logic;
signal dcounter_en                : std_logic;
signal data_end_count             : std_logic_vector(23 downto 0);
signal endcnt                     : std_logic_vector(31 downto 0);
signal maxendcnt                  : std_logic;
signal test_run                   : std_logic_vector(3 downto 0); -- up to 16 types of data/clk schemes
signal timer                      : std_logic_vector(31 downto 0); 
signal timer_cnten                : std_logic;
signal timer_clr                  : std_logic;
signal idle_flag                  : std_logic;
signal start_flag                 : std_logic;
signal eye_flag                   : std_logic;
signal run_flag                   : std_logic;
signal endrun1_flag               : std_logic;
signal endrun_flag                : std_logic;
signal test_fail_flag             : std_logic;
signal test_success_flag          : std_logic;
signal test_fail_flag_r           : std_logic;
signal test_success_flag_r        : std_logic;
signal fiforst_flag               : std_logic;
signal control_reg                : std_logic_vector(47 downto 0);
signal control_start_bit          : std_logic;
signal control_smreset_bit        : std_logic;
signal control_dut_rst_bit        : std_logic;
signal control_dcount_disable_bit      : std_logic;
signal idelay_reg                 : std_logic_vector(47 downto 0);
signal idelay_d0_bfield           : std_logic_vector(4 downto 0);
signal idelay_d1_bfield           : std_logic_vector(4 downto 0);
signal idelay_d2_bfield           : std_logic_vector(4 downto 0);
signal idelay_d3_bfield           : std_logic_vector(4 downto 0);
signal uart_start_stb             : std_logic;
signal to_error_det               : std_logic;
signal sink_fifo_wren_1           : std_logic;
signal sink_lfsr_data_r           : std_logic_vector(31 downto 0);
signal sink_fifo_dout_r           : std_logic_vector(31 downto 0);
signal debug_smstates             : std_logic_vector(3 downto 0);
signal test_done_flag             : std_logic;
signal xadc_di                    : std_logic_vector(15 downto 0);
signal xadc_do                    : std_logic_vector(15 downto 0);
signal xadc_dwe                   : std_logic;
signal xadc_drdy                  : std_logic;
signal xadc_den                   : std_logic;
signal xadc_addr                  : std_logic_vector(6 downto 0);
signal xadc_alarm                 : std_logic_vector(7 downto 0);
signal xadc_overtemp_alarm        : std_logic;
signal xadc_we_ctrl_bit           : std_logic;
signal fan_on                     : std_logic;

-------------------------------------------------------------------------------
--  Procedure Declarations
-------------------------------------------------------------------------------
  
  
begin


-------------------------------------------------------------------------------
--  Component Instantiations
-------------------------------------------------------------------------------


  GEN_Si570_g2 : if IS_GENESYS2_BOARD = 1 generate
    Si570_inst : Genesys2_Si570_INF 
    port map(
          clk_100                    => tb_clock_100,
          areset                     => areset,
          smreset                    => '0',
          clk100_heartbeat_led       => tb_clock_100_heartbeat,
          clk_si570_heartbeat_led    => dut_clk_heartbeat,
          si570_div16_mon            => si570_div16_monitor,
          reg_address                => ireg_address,
          reg_data_out               => ireg_data_out,
          reg_write_stb              => ireg_write_stb,
          reg_data_in                => ireg_data_in,
          reg_read_stb               => ireg_read_stb,
          busy                       => ibusy,
          scl                        => scl_i,
          sda_in                     => sda_in_i,
          sda_out                    => sda_out_i,
          sda_oe                     => sda_oe_i,
          uart_rx                    => uart_rx_out,
          uart_tx                    => uart_tx_in,
          vclk                       => dut_clk);
  end generate;

  GEN_Si570_nv : if IS_GENESYS2_BOARD = 0 generate
    Si570_inst : NexysVideo_Si570_INF 
    port map(
          clk_100                    => tb_clock_100,
          areset                     => areset,
          smreset                    => '0',
          clk100_heartbeat_led       => tb_clock_100_heartbeat,
          clk_si570_heartbeat_led    => dut_clk_heartbeat,
          si570_div16_mon            => si570_div16_monitor,
          reg_address                => ireg_address,
          reg_data_out               => ireg_data_out,
          reg_write_stb              => ireg_write_stb,
          reg_data_in                => ireg_data_in,
          reg_read_stb               => ireg_read_stb,
          busy                       => ibusy,
          scl                        => scl_i,
          sda_in                     => sda_in_i,
          sda_out                    => sda_out_i,
          sda_oe                     => sda_oe_i,
          uart_rx                    => uart_rx_out,
          uart_tx                    => uart_tx_in,
          vclk                       => dut_clk);
  end generate;

 
  -- Testbed to Data Source interface 512x36-bit  
    
  -- FIFO18E1: 18Kb FIFO (First-In-First-Out) Block RAM Memory
  -- 7 Series
  -- Xilinx HDL Libraries Guide, version 13.4
  SourceFifo_inst : FIFO18E1
  generic map (
    ALMOST_EMPTY_OFFSET       => X"0010",       -- Sets the almost empty threshold
    ALMOST_FULL_OFFSET        => X"0010",       -- Sets almost full threshold
    DATA_WIDTH                => 36,            -- Sets data width to 4-36
    DO_REG                    => 1,             -- Enable output register (1-0) Must be 1 if EN_SYN = FALSE
    EN_SYN                    => FALSE,          -- Specifies FIFO as dual-clock (FALSE) or Synchronous (TRUE)
    FIFO_MODE                 => "FIFO18_36",   -- Sets mode to FIFO18 or FIFO18_36
    FIRST_WORD_FALL_THROUGH   => FALSE,         -- Sets the FIFO FWFT to FALSE, TRUE
    INIT                      => X"000000000",  -- Initial values on output port
    SIM_DEVICE                => "7SERIES",     -- Must be set to "7SERIES" for simulation behavior
    SRVAL                     => X"000000000"   -- Set/Reset value for output port
  )
  port map (
    -- Read Data: 32-bit (each) output: Read output data
    DO              => source_fifo_data, -- 32-bit output: Data output
    DOP             => source_fifo_pdata, -- 4-bit output: Parity data output
    -- Status: 1-bit (each) output: Flags and other FIFO status outputs
    ALMOSTEMPTY     => open, -- 1-bit output: Almost empty flag
    ALMOSTFULL      => open, -- 1-bit output: Almost full flag
    EMPTY           => source_fifo_empty_flag, -- 1-bit output: Empty flag
    FULL            => open, -- 1-bit output: Full flag
    RDCOUNT         => open, -- 12-bit output: Read count
    RDERR           => open, -- 1-bit output: Read error
    WRCOUNT         => source_fifo_write_count, -- 12-bit output: Write count
    WRERR           => open, -- 1-bit output: Write error occured.
    -- Read Control Signals: 1-bit (each) input: Read clock, enable and reset input signals
    RDCLK           => dut_clk, -- 1-bit input: Read clock
    RDEN            => source_fifo_rd, -- 1-bit input: Read enable
    REGCE           => '1', -- 1-bit input: Clock enable
    RST             => source_fifo_reset, -- 1-bit input: Asyncronous Reset
    RSTREG          => '0', -- 1-bit input: Output register set/reset
    -- Write Control Signals: 1-bit (each) input: Write clock and enable input signals
    WRCLK           => tb_clock_100, -- 1-bit input: Write clock
    WREN            => source_fifo_wr, -- 1-bit input: Write enable
    -- Write Data: 32-bit (each) input: Write input data
    DI              => source_fifo_din, -- 32-bit input: Data input 
    DIP             => source_fifo_pdin -- 4-bit input: Parity input
  );
 
  -- Testbed to Data Sink interface  
    
  -- FIFO18E1: 18Kb FIFO (First-In-First-Out) Block RAM Memory
  -- 7 Series
  -- Xilinx HDL Libraries Guide, version 13.4
  SinkFifo_inst : FIFO18E1
  generic map (
    ALMOST_EMPTY_OFFSET       => X"0010",       -- Sets the almost empty threshold
    ALMOST_FULL_OFFSET        => X"0010",       -- Sets almost full threshold
    DATA_WIDTH                => 36,            -- Sets data width to 4-36
    DO_REG                    => 1,             -- Enable output register (1-0) Must be 1 if EN_SYN = FALSE
    EN_SYN                    => FALSE,          -- Specifies FIFO as dual-clock (FALSE) or Synchronous (TRUE)
    FIFO_MODE                 => "FIFO18_36",   -- Sets mode to FIFO18 or FIFO18_36
    FIRST_WORD_FALL_THROUGH   => FALSE,         -- Sets the FIFO FWFT to FALSE, TRUE
    INIT                      => X"000000000",  -- Initial values on output port
    SIM_DEVICE                => "7SERIES",     -- Must be set to "7SERIES" for simulation behavior
    SRVAL                     => X"000000000"   -- Set/Reset value for output port
  )
  port map (
    -- Read Data: 32-bit (each) output: Read output data
    DO              => sink_fifo_dout, -- 32-bit output: Data output
    DOP             => sink_fifo_pdout, -- 4-bit output: Parity data output
    -- Status: 1-bit (each) output: Flags and other FIFO status outputs
    ALMOSTEMPTY     => open, -- 1-bit output: Almost empty flag
    ALMOSTFULL      => open, -- 1-bit output: Almost full flag
    EMPTY           => sink_fifo_empty, -- 1-bit output: Empty flag
    FULL            => sink_fifo_full_flag, -- 1-bit output: Full flag
    RDCOUNT         => open, -- 12-bit output: Read count
    RDERR           => open, -- 1-bit output: Read error
    WRCOUNT         => open, -- 12-bit output: Write count
    WRERR           => open, -- 1-bit output: Write error occured.
    -- Read Control Signals: 1-bit (each) input: Read clock, enable and reset input signals
    RDCLK           => tb_clock_100, -- 1-bit input: Read clock
    RDEN            => sink_fifo_rd, -- 1-bit input: Read enable
    REGCE           => '1', -- 1-bit input: Clock enable
    RST             => sink_fifo_reset, -- 1-bit input: Asyncronous Reset
    RSTREG          => '0', -- 1-bit input: Output register set/reset
    -- Write Control Signals: 1-bit (each) input: Write clock and enable input signals
    WRCLK           => sample_clk, -- 1-bit input: Write clock      !
    WREN            => sink_fifo_wren_1, -- 1-bit input: Write enable
    -- Write Data: 32-bit (each) input: Write input data
    DI              => sink_fifo_data, -- 32-bit input: Data input
    DIP             => sink_fifo_pdata -- 4-bit input: Parity input
  );

  -- XADC: Dual 12-Bit 1MSPS Analog-to-Digital Converter
  -- 7 Series
  -- Xilinx HDL Libraries Guide, version 13.4
  XADC_inst : XADC
  generic map (
    -- INIT_40 - INIT_42: XADC configuration registers
    INIT_40 => X"A000", -- 64 sample averaging enabled
    INIT_41 => X"2ef0",
    INIT_42 => X"0800", -- ADCCLK = 100/8 = 12.5 Mhz
    -- INIT_48 - INIT_4F: Sequence Registers
    INIT_48 => X"4701", -- CHSEL1 - enable Temp VCCINT, VCCAUX, VCCBRAM, and calibration
    INIT_49 => X"0000", -- CHSEL2 - enable nothing else
    INIT_4A => X"0001", -- SEQAVG1 disabled all channels
    INIT_4B => X"0000", -- SEQAVG2 disabled all channels
    INIT_4C => X"0000", -- SEQINMODE0 - all channels unipolar
    INIT_4D => X"0000", -- SEQINMODE1 - all channels unipolar
    INIT_4E => X"0000", -- SEQACQ0 - No extra settling time all channels
    INIT_4F => X"0000", -- SEQACQ1 - No extra settling time all channels
    -- INIT_50 - INIT_58, INIT5C: Alarm Limit Registers
    INIT_50 => X"b5ed", -- Temp upper alarm trigger 85C
    INIT_51 => X"5999", -- Vccint upper alarm limit 1.05V
    INIT_52 => X"A147", -- Vccaux upper alarm limit 1.89V
    INIT_53 => X"dddd", -- OT upper alarm limit 125C - see Thermal Management
    INIT_54 => X"a93a", -- Temp lower alarm reset 60C
    INIT_55 => X"5111", -- Vccint lower alarm limit 0.95V
    INIT_56 => X"91Eb", -- Vccaux lower alarm limit 1.71V
    INIT_57 => X"ae4e", -- OT lower alarm reset 70C - see Thermal Management
    INIT_58 => X"5999", -- VCCBRAM upper alarm limit 1.05V
    INIT_5C => X"5111", -- VCCBRAM lower alarm limit 0.95V
    -- Simulation attributes: Set for proepr simulation behavior
    SIM_DEVICE => "7SERIES", -- Select target device (values)
    SIM_MONITOR_FILE => "design.txt" -- Analog simulation data file name
  )
  port map (
    -- ALARMS: 8-bit (each) output: ALM, OT
    ALM             => xadc_alarm,          -- 8-bit output: Output alarm for temp, Vccint, Vccaux and Vccbram
    OT              => xadc_overtemp_alarm, -- 1-bit output: Over-Temperature alarm
    -- Dynamic Reconfiguration Port (DRP): 16-bit (each) output: Dynamic Reconfiguration Ports
    DO              => xadc_do,             -- 16-bit output: DRP output data bus
    DRDY            => xadc_drdy,           -- 1-bit output: DRP data ready
    -- STATUS: 1-bit (each) output: XADC status ports
    BUSY            => open, -- 1-bit output: ADC busy output
    CHANNEL         => open, -- 5-bit output: Channel selection outputs
    EOC             => open, -- 1-bit output: End of Conversion
    EOS             => open, -- 1-bit output: End of Sequence
    JTAGBUSY        => open, -- 1-bit output: JTAG DRP transaction in progress output
    JTAGLOCKED      => open, -- 1-bit output: JTAG requested DRP port lock
    JTAGMODIFIED    => open, -- 1-bit output: JTAG Write to the DRP has occurred
    MUXADDR         => open, -- 5-bit output: External MUX channel decode
    -- Auxiliary Analog-Input Pairs: 16-bit (each) input: VAUXP[15:0], VAUXN[15:0]
    VAUXN           => (others => '0'),     -- 16-bit input: N-side auxiliary analog input
    VAUXP           => (others => '0'),     -- 16-bit input: P-side auxiliary analog input
    -- CONTROL and CLOCK: 1-bit (each) input: Reset, conversion start and clock inputs
    CONVST          => '0',                 -- 1-bit input: Convert start input
    CONVSTCLK       => '0',                 -- 1-bit input: Convert start input
    RESET           => '0',                 -- 1-bit input: Active-high reset
    -- Dedicated Analog Input Pair: 1-bit (each) input: VP/VN
    VN              => '0',                 -- 1-bit input: N-side analog input
    VP              => '0',                 -- 1-bit input: P-side analog input
    -- Dynamic Reconfiguration Port (DRP): 7-bit (each) input: Dynamic Reconfiguration Ports
    DADDR           => xadc_addr,           -- 7-bit input: DRP address bus
    DCLK            => tb_clock_100,        -- 1-bit input: DRP clock
    DEN             => xadc_den,            -- 1-bit input: DRP enable signal
    DI              => xadc_di,             -- 16-bit input: DRP input data bus
    DWE             => xadc_dwe             -- 1-bit input: DRP write enable
  );
  
  
-------------------------------------------------------------------------------
--  Concurrent Assignments
-------------------------------------------------------------------------------

  sreset    <= control_smreset_bit;
   
  -- These are the debug registers available to this entity
  ireg_data_in  <= dbg_reg0_r     when ireg_address = X"0" else
                   dbg_reg1_r     when ireg_address = X"1" else
                   dbg_reg2_r     when ireg_address = X"2" else
                   dbg_reg3_r     when ireg_address = X"3" else
                   dbg_reg4_r     when ireg_address = X"4" else
                   dbg_reg5_r     when ireg_address = X"5" else
                   dbg_reg6_r     when ireg_address = X"6" else
                   dbg_reg7_r     when ireg_address = X"7" else
                   dbg_reg8_r     when ireg_address = X"8" else
                   dbg_reg9_r     when ireg_address = X"9" else
                   dbg_reg10_r    when ireg_address = X"A" else
                   dbg_reg11_r    when ireg_address = X"B" else
                   -- registers 12 through 15 are used by Si570_INF
                   (others => '0');
         
  idle_flag            <= '1' when current_state = IDLE     else '0';       
  start_flag           <= '1' when current_state = START    else '0';       
  run_flag             <= '1' when current_state = RUN      else '0';
  eye_flag             <= '1' when current_state = EYE      else '0';      
  endrun1_flag         <= '1' when current_state = ENDRUN1  else '0';
  endrun_flag          <= '1' when current_state = ENDRUN   else '0';      
  test_fail_flag       <= '1' when current_state = OHCRAP   else '0';             
  test_success_flag    <= '1' when current_state = ENDRUN   else '0';
  fiforst_flag         <= '1' when current_state = FIFO_RST else '0'; 
  
  -- not to useful except with and ILA or UART and state machine IDLE or stuck
  debug_smstates        <= X"0" when  current_state = IDLE        else 
                           X"1" when  current_state = FIFO_RST    else 
                           X"2" when  current_state = START       else 
                           X"3" when  current_state = RUN         else 
                           X"4" when  current_state = OHCRAP      else 
                           X"5" when  current_state = ENDRUN1     else 
                           X"6" when  current_state = ENDRUN      else 
                           X"F";

  source_lfsr_en       <= source_fifo_wr;
  source_lfsr_ld       <= '1' when idle_flag = '1' else '0';
                     
  source_fifo_reset   <= '1' when idle_flag = '1' else '0';
  sink_fifo_reset     <= '1' when idle_flag = '1' else '0'; 
  
  sink_fifo_wren_1    <= '0' when idle_flag = '1' else
                         '0' when fiforst_flag = '1' else 
                         sink_fifo_wren ;                   
                       
  source_fifo_wr      <= '0' when idle_flag = '1' else
                         '0' when fiforst_flag = '1' else  
                         '1' when start_flag = '1' else                       
                         source_fifo_empty_tb_clock_100_r;
   
  source_fifo_rd      <= '0' when idle_flag = '1' else
                         '0' when fiforst_flag = '1' else  -- observe series 7 hardware fifo rules 
                         '0' when start_flag = '1' else 
                         source_fifo_rden;
  

  source_fifo_din   <= source_lfsr_data; 
  source_fifo_pdin  <= "000" & run_enable;

  sink_lfsr_en      <= sink_fifo_rd;
  sink_lfsr_ld      <= '1' when idle_flag = '1' else '0';
  

  sink_fifo_rd      <= '1' when  run_flag = '1' and sink_fifo_empty = '0' else '0'; -- observe series 7 hardware fifo rules  
    
  
  start_test        <= '1' when simulation_mode_flag = '1' and sim_start_stb = '1' else
                       '1' when uart_start_stb = '1' else
                       '0';
                   
  lfsr_seed         <= sim_source_lfsr_seed when simulation_mode_flag = '1' else  source_lfsr_seed;             
   
  data_end_count  <= sim_run_words_m1 when simulation_mode_flag = '1' else run_words_m1;
                   
  test_complete   <= '1' when data_counter > data_end_count else '0';
                   
  run_enable      <= '1' when run_flag = '1' else
                     '1' when eye_flag = '1' else
                     '1' when start_flag = '1' else
                     '0'; 
   
  dcounter_ld     <= '1' when start_flag = '1' else
                     '0';
                     
  dcounter_en     <= '1' when run_flag = '1' and test_complete = '0' and sink_lfsr_en = '1' and control_dcount_disable_bit = '0' else
                     '0';
  
  maxendcnt       <= '1' when fiforst_flag = '1' and endcnt > X"000000FF" else
                     '1' when endrun1_flag = '1' and endcnt > X"000000FF" else 
                     '0';  -- 16 us                                    

  sda_in_i        <= sda_in;

  timer_cnten     <= '1' when run_flag = '1' and test_complete = '0' else '0';
  timer_clr       <= '1' when start_flag = '1' else '0';
  
-------------------------------------------------------------------------------
--  Process Declarations
-------------------------------------------------------------------------------


-- The UART DEBUG registers
  -- the low 12 registers can be used as read-back, read-write, read-only or write-only
  -- by the top level
  -- For this testbed we'll assign r/w usage to the following registers: 
  --  R0 as a read-back to test the interface. Writable by the UART. Readable by this entity
  --  R1 as a test ID. Writable by this entity. Readable by the UART
  --  R2 Test Status
  --  R3 Test Results. Writable by either this entity or the UART
  --  R4 Test Data Sink word count read counter
  --  R8 Test control register  
  
  -- dbg_reg1_r is static so we can just make a concurrent assignment
  dbg_reg1_r      <= test_id;
  
  -- R2 Status Register
  -- Writeable by the UART or this entity.The UART has write precidence.
  -- R2 is written here when the test state machine hits either the OHCRAP
  -- or SUCCESS states. Either way it writes the folowing.
  -- R2(0)    - test complete flag
  -- R2 can be cleared but not written by the UART 
  reg2_wr_data    <= debug_smstates & X"0000000000" & "000" & test_done_flag;                 
  reg2_wr_stb     <= '1' when ireg_write_stb = '0' else '0';  
  TEST_STATUS_REG2_WRITE_PROC : process(tb_clock_100, areset)
  begin
    if areset = '1' then
      dbg_reg2_r      <= (others => '0');
    elsif rising_edge(tb_clock_100) then
      if (ireg_write_stb = '1' and ireg_address = X"2") then
        dbg_reg2_r   <= (others => '0');
      elsif reg2_wr_stb = '1' then  
        dbg_reg2_r   <= reg2_wr_data;
      else
        dbg_reg2_r   <= dbg_reg2_r;
      end if;
    end if;
  end process TEST_STATUS_REG2_WRITE_PROC;

  -- R3 Test Result Register
  -- Writeable by the UART or this entity.The UART has write precidence.
  -- R3 is written here when the test state machine hits the OHCRAP
  -- or SUCCESS states. Either way it writes the folowing.
  -- R3(47)   - error detection status
  -- R3(46)   - timeout error detection status
  -- R3(3:0)  - data sink control bits
  -- R3 can be cleared but not written by the UART 
  reg3_wr_data    <= error_det & to_error_det & "00" & X"0000000000"& sink_fifo_pdata;                 
  reg3_wr_stb     <= '1' when ireg_write_stb = '0' else '0';  
  TEST_RESULT_REG3_WRITE_PROC : process(tb_clock_100, areset)
  begin
    if areset = '1' then
      dbg_reg3_r      <= (others => '0');
    elsif rising_edge(tb_clock_100) then
      if (ireg_write_stb = '1' and ireg_address = X"3") then
        dbg_reg3_r   <= (others => '0');
      elsif reg3_wr_stb = '1' then  
        dbg_reg3_r   <= reg3_wr_data;
      else
        dbg_reg3_r   <= dbg_reg3_r;
      end if;
    end if;
  end process TEST_RESULT_REG3_WRITE_PROC;
  

  -- R4 Test Result Register
  -- Writeable by the UART or this entity.The UART has write precidence.
  -- R4 is written here when the test state machine hits the OHCRAP
  -- or SUCCESS states. Either way it writes the folowing.
  -- R4   - data sink word count
  -- R4 can be cleared but not written by the UART 
  reg4_wr_data    <= X"000000" & data_counter;                 
  reg4_wr_stb     <= '1' when ireg_write_stb = '0' else '0';  
  WORD_COUNT_REG4_WRITE_PROC : process(tb_clock_100, areset)
  begin
    if areset = '1' then
      dbg_reg4_r      <= (others => '0');
    elsif rising_edge(tb_clock_100) then
      if (ireg_write_stb = '1' and ireg_address = X"4") then
        dbg_reg4_r   <= (others => '0');
      elsif reg4_wr_stb = '1' then  
        dbg_reg4_r   <= reg4_wr_data;
      else
        dbg_reg4_r   <= dbg_reg4_r;
      end if;
    end if;
  end process WORD_COUNT_REG4_WRITE_PROC;

  -- R5 Test Result Register
  -- Writeable by the UART or this entity.The UART has write precidence.
  -- R5 is written here when the test state machine hits the OHCRAP
  -- or SUCCESS states. Either way it writes the folowing.
  -- R5  - Elapsed time for the test run
  -- R5 can be cleared but not written by the UART 
  reg5_wr_data    <= X"0000" & timer;                 
  reg5_wr_stb     <= '1' when ireg_write_stb = '0' else '0';                      
  ELAPSED_TIME_REG5_WRITE_PROC : process(tb_clock_100, areset)
  begin
    if areset = '1' then
      dbg_reg5_r      <= (others => '0');
    elsif rising_edge(tb_clock_100) then
      if (ireg_write_stb = '1' and ireg_address = X"5") then
        dbg_reg5_r   <= (others => '0');
      elsif reg5_wr_stb = '1' then  
        dbg_reg5_r   <= reg5_wr_data;
      else
        dbg_reg5_r   <= dbg_reg5_r;
      end if;
    end if;
  end process ELAPSED_TIME_REG5_WRITE_PROC;
  
  -- R8 is the control register. Writable by the UART                   
  control_reg         <= dbg_reg8_r;    
  
  -- Control Register Bit assignments                    
  control_start_bit               <= control_reg(0);  -- UART can start a test run by setting this to 1                   
  control_dcount_disable_bit      <= control_reg(1);  -- UART can run EYE mode by setting this to 1 ( RUN never completes )
  control_dut_rst_bit             <= control_reg(2);  -- UART can reset the data source and data sink
  control_smreset_bit             <= control_reg(3);  -- UART can force test state machine into IDLE mode
  
  
  -- read data sink data and sink_lfsr data for debugging errors
  dbg_reg6_r    <= X"0000" & sink_lfsr_data_r;
  dbg_reg7_r    <= X"0000" & sink_fifo_dout_r;
         
  
  -- R9 is the idelay register. Writable by the UART                   
  idelay_reg          <= dbg_reg9_r;
  idelay_d0_bfield    <= idelay_reg(4 downto 0);    -- UART can set data sink data bit 0 idelay
  idelay_d1_bfield    <= idelay_reg(12 downto 8);   -- UART can set data sink data bit 1 idelay  
  idelay_d2_bfield    <= idelay_reg(20 downto 16);  -- UART can set data sink data bit 2 idelay
  idelay_d3_bfield    <= idelay_reg(28 downto 24);  -- UART can set data sink data bit 3 idelay

  -- R10 is the XADC Write Register. Writable by the UART
  xadc_we_ctrl_bit    <= dbg_reg10_r(47);
  xadc_addr           <= dbg_reg10_r(22 downto 16);
  xadc_di             <= dbg_reg10_r(15 downto 0);
  
  xadc_den            <= '1' when ireg_write_stb = '1' and ireg_address = X"A" else '0';
   
  -- In XADC DRP timing xadc_dwe and xadc_di is sampled on the rising_edge of DCLK following 
  -- assertion of xadc_den 
  process(tb_clock_100,areset)
  begin
    if areset = '1' then
      xadc_dwe <= '0';
    elsif (rising_edge(tb_clock_100)) then
      xadc_dwe <= xadc_den and xadc_we_ctrl_bit;
    end if;                                                                       
  end process ; 
  
  -- R11 XADC Read Register
  -- Writeable by the UART or this entity.The UART has write precidence.
  -- R11 is written here when the the XADC DRDY strobe is asserted
  -- R11   - XADC DRP DO output
  -- R11 can be cleared but not written by the UART 
  reg11_wr_data    <= xadc_overtemp_alarm & xadc_alarm & X"00000" & "000" & xadc_do;                 
  reg11_wr_stb     <= '1' when xadc_drdy = '1' else '0';  
  XADC_READ_REG11_WRITE_PROC : process(tb_clock_100, areset)
  begin
    if areset = '1' then
      dbg_reg11_r      <= (others => '0');
    elsif rising_edge(tb_clock_100) then
      if (ireg_write_stb = '1' and ireg_address = X"B") then
        dbg_reg11_r   <= (others => '0');
      elsif reg11_wr_stb = '1' then  
        dbg_reg11_r   <= reg11_wr_data;
      else
        dbg_reg11_r   <= dbg_reg11_r;
      end if;
    end if;
  end process XADC_READ_REG11_WRITE_PROC;
        
  -- a process that allows the UART to write read-back registers
  UART_DBG_REG_WRITE_PROC : process(tb_clock_100, areset)
  begin
    if areset = '1' then
      dbg_reg0_r      <= (others => '0');
      dbg_reg8_r      <= (others => '0');
      dbg_reg9_r      <= (others => '0');
      dbg_reg10_r     <= (others => '0');
    elsif rising_edge(tb_clock_100) then
      if (ireg_write_stb = '1') then
        if ireg_address = X"0" then
          dbg_reg0_r   <= ireg_data_out;
        elsif ireg_address = X"8" then
          dbg_reg8_r   <= ireg_data_out;
        elsif ireg_address = X"9" then
          dbg_reg9_r   <= ireg_data_out;
        elsif ireg_address = X"A" then
          dbg_reg10_r   <= ireg_data_out;
        -- registers 12 through 15 are used by Si570_INF           
        end if;
      else
        dbg_reg0_r      <= dbg_reg0_r;
        dbg_reg8_r      <= dbg_reg8_r;
        dbg_reg9_r      <= dbg_reg9_r;
        dbg_reg10_r     <= dbg_reg10_r;
        -- registers 12 through 15 are used by Si570_INF  
      end if;
    end if;
  end process UART_DBG_REG_WRITE_PROC;
                                                                 
  -- a 32-bit pseudo-random word generator
  SRC_LFSR_32: process(tb_clock_100)
  begin
    if rising_edge(tb_clock_100) then
      if source_lfsr_ld = '1' then
        source_lfsr_data  <= lfsr_seed;
      elsif source_lfsr_en = '1' then
        source_lfsr_data(31 downto 1)  <= source_lfsr_data(30 downto 0);
        source_lfsr_data(0) <= source_lfsr_data(31) xor source_lfsr_data(6) xor source_lfsr_data(5) xor source_lfsr_data(1);
      end if;
    end if;
  end process SRC_LFSR_32;

                   
-- sink data check generator --------------------------
                                                                 
  -- a 32-bit pseudo-random word generator
  SINK_LFSR_32: process(tb_clock_100)
  begin
    if rising_edge(tb_clock_100) then
      if sink_lfsr_ld = '1' then
        sink_lfsr_data  <= lfsr_seed;
      elsif sink_lfsr_en_r4 = '1' then
        sink_lfsr_data(31 downto 1)  <= sink_lfsr_data(30 downto 0);
        sink_lfsr_data(0) <= sink_lfsr_data(31) xor sink_lfsr_data(6) xor sink_lfsr_data(5) xor sink_lfsr_data(1);
      end if;
    end if;
  end process SINK_LFSR_32;
  
  -- keep track of words written by data sink while the test is running
  test_data_count_proc : process(tb_clock_100,areset)
  begin
    if areset = '1' then
      data_counter <= (others => '0');
    elsif (rising_edge(tb_clock_100)) then
      if dcounter_ld = '1' then            
        data_counter <= (others => '0');
      elsif dcounter_en = '1' then            
        data_counter <= data_counter + 1;
      else  
        data_counter <= data_counter;
      end if;  
    end if;                                                                       
  end process test_data_count_proc;  
  
  -- data sink error checker
  test_data_checker_proc : process(tb_clock_100,areset)
  begin
    if areset = '1' then
      error_det   <= '0';
    elsif (rising_edge(tb_clock_100)) then
      if fiforst_flag = '1' then            
        error_det <= '0';
      elsif run_flag = '1' and sink_lfsr_en_r2 = '1' and sink_lfsr_data /= sink_fifo_dout then            
        error_det <= '1';
      else  
        error_det <= error_det;
      end if;  
    end if;                                                                       
  end process test_data_checker_proc;  

  
  process(tb_clock_100,areset)
  begin
    if areset = '1' then
      sink_lfsr_data_r <= (others => '0');
      sink_fifo_dout_r <= (others => '0');
    elsif (rising_edge(tb_clock_100)) then
      if sink_lfsr_en_r1 = '1' then
        sink_lfsr_data_r  <= sink_lfsr_data;
        sink_fifo_dout_r  <= sink_fifo_dout;
      elsif start_flag = '1' then
        sink_lfsr_data_r <= (others => '0');
        sink_fifo_dout_r <= (others => '0');
      else  
        sink_lfsr_data_r <= sink_lfsr_data_r;
        sink_fifo_dout_r <= sink_fifo_dout_r;
      end if;
    end if;                                                                       
  end process ; 
    
  process(tb_clock_100,areset)
  begin
    if areset = '1' then
      sink_lfsr_en_r1 <= '0';
      sink_lfsr_en_r2 <= '0';
      sink_lfsr_en_r3 <= '0';
      sink_lfsr_en_r4 <= '0';
    elsif (rising_edge(tb_clock_100)) then
      sink_lfsr_en_r1 <= sink_lfsr_en;
      sink_lfsr_en_r2 <= sink_lfsr_en_r1;
      sink_lfsr_en_r3 <= sink_lfsr_en_r2;
      sink_lfsr_en_r4 <= sink_lfsr_en_r3;
    end if;                                                                       
  end process ; 
   
  -- get the source write fifo empty flag into the tb_clock_100 domain
  process(tb_clock_100,areset)
  begin
    if areset = '1' then
      source_fifo_empty_tb_clock_100 <= '0';
      source_fifo_empty_tb_clock_100_r <= '0';
    elsif (rising_edge(tb_clock_100)) then
      source_fifo_empty_tb_clock_100 <= source_fifo_empty_flag;
      source_fifo_empty_tb_clock_100_r <= source_fifo_empty_tb_clock_100;
    end if;                                                                       
  end process ; 
  
  process(tb_clock_100,areset)
  begin
    if areset = '1' then
      test_done_flag <= '0';
    elsif (rising_edge(tb_clock_100)) then
      if start_flag = '1' then
        test_done_flag <= '0';
      elsif ireg_write_stb = '1' and ireg_address = X"2" then  -- let UART clear this flag
        test_done_flag <= '0';
      elsif test_fail_flag_r = '1' then
        test_done_flag <= '1';
      elsif test_success_flag_r = '1' then
        test_done_flag <= '1';
      end if;
    end if;                                                                       
  end process ; 
    
  -- a delay timer to allow data sources and sinks to terminate gracefully
  -- reused to observe fifo reset requirements
  testend_count_proc : process(tb_clock_100,areset)
  begin
    if areset = '1' then
      endcnt <= (others => '0');
    elsif (rising_edge(tb_clock_100)) then
      if endrun1_flag = '1' and maxendcnt = '0' then 
        endcnt <= endcnt + 1;
      elsif fiforst_flag = '1' and maxendcnt = '0' then  
        endcnt <= endcnt + 1;
      elsif idle_flag = '1' then  
        endcnt <= (others => '0');
      else  
        endcnt <= endcnt;
      end if;  
    end if;                                                                       
  end process testend_count_proc;                                                    
     
  -- a timer to measure the elapsed time of test runs
  test_timer_proc : process(tb_clock_100,areset)
  begin
    if areset = '1' then
      timer <= (others => '0');
    elsif (rising_edge(tb_clock_100)) then
      if timer_clr = '1' then  
        timer <= (others => '0');
      elsif timer_cnten = '1' then 
        timer <= timer + 1;
      else  
        timer <= timer;
      end if;  
    end if;                                                                       
  end process test_timer_proc; 
  
  -- uart_start_stb
  -- only gets asserted when test_complete bit in R32 is 0
  process(tb_clock_100,areset)
  begin
    if areset = '1' then
      uart_start_stb <= '0';
    elsif (rising_edge(tb_clock_100)) then
      if control_start_bit = '1' and idle_flag = '1' then
        uart_start_stb <= '1';
      elsif control_start_bit = '0' then
        uart_start_stb <= '0';
      end if;
    end if;                                                                       
  end process ; 
  
  
  -- time-out error checker ( about 1 ms without a data sink fifo write )
  -- set if timer(17) = '1' and data_counter < 1
  timeout_checker_proc : process(tb_clock_100,areset)
  begin
    if areset = '1' then
      to_error_det   <= '0';
    elsif (rising_edge(tb_clock_100)) then
      if start_flag = '1' then            
        to_error_det <= '0';
      elsif timer(17) = '1' and data_counter < X"00000F" then            
        to_error_det <= '1';
      else  
        to_error_det <= to_error_det;
      end if;  
    end if;                                                                       
  end process timeout_checker_proc;  
  
  process(tb_clock_100,areset)
  begin
    if areset = '1' then
      test_fail_flag_r <= '0';
    elsif (rising_edge(tb_clock_100)) then
      if start_flag = '1' then
        test_fail_flag_r <= '0';
      elsif test_fail_flag = '1' then
        test_fail_flag_r <= '1';
      end if;
    end if;                                                                       
  end process ; 
  
  process(tb_clock_100,areset)
  begin
    if areset = '1' then
      test_success_flag_r <= '0';
    elsif (rising_edge(tb_clock_100)) then
      if start_flag = '1' then
        test_success_flag_r <= '0';
      elsif test_success_flag = '1' then
        test_success_flag_r <= '1';
      end if;
    end if;                                                                       
  end process ; 
  
  -- xadc based fan control
  process(tb_clock_100,areset)
  begin
    if areset = '1' then
      fan_on <= '0';
    elsif (rising_edge(tb_clock_100)) then
--      if dbg_reg11_r(15 downto 0) > XADC_FAN_ON_TEMP then -- use this if you don't want to hear the fan
      if dbg_reg11_r(15 downto 0) > XADC_38_C_FAN_ON_TEMP then
        fan_on <= '1';
      elsif dbg_reg11_r(15 downto 0) < XADC_35_C_FAN_OFF_TEMP then
        fan_on <= '0';
      else
        fan_on  <= fan_on;
      end if;
    end if;                                                                       
  end process ; 
  
                                                  

-- a two-processs state machine for to run the test
  
  -- the clocked portion
  UPDATE_STATE : process(tb_clock_100, areset)
  begin
    if areset = '1' then
      current_state   <= IDLE;
    elsif rising_edge(tb_clock_100) then
      if sreset = '1' then
        current_state   <= IDLE;
      else  
        current_state   <= next_state;
      end if;  
    end if;
  end process UPDATE_STATE;

  -- the state machine
  -- type sm_states is (IDLE, FIFO_RST, START, RUN, EYE, OHCRAP, ENDRUN, ENDRUN1);
  
  TEST_CONTROL_SM : process (current_state, start_test, error_det, test_complete, maxendcnt, to_error_det, control_dcount_disable_bit)           
  begin                                                       
    case current_state is
      when IDLE =>
        if start_test = '1' then
          next_state  <= FIFO_RST;
        else
          next_state  <= IDLE;
        end if;
        
      -- ensure that fifo reset is released 2 clocks before fifo_we is asserted  
      when FIFO_RST =>
        if maxendcnt = '1' then
          next_state  <= START;
        else
          next_state  <= FIFO_RST;
        end if;
        
      -- this prepares the testbed for a test run  
      when START =>
        if control_dcount_disable_bit = '1' then
          next_state  <= EYE;
        else 
          next_state  <= RUN;  
        end if;

      -- stay in this mode until sreset  
      when EYE =>
        next_state  <= EYE;  
                
      -- run until an error is detected  
      when RUN =>
        if error_det = '1' then
          next_state  <= OHCRAP;
        elsif to_error_det = '1' then
          next_state  <= OHCRAP;
        elsif test_complete = '1' then
          next_state  <= ENDRUN1;
        else
          next_state  <= RUN;
        end if;
                
      -- OOPS! The data sink and data source don't agree with each other             
      -- make sure that start_test is de-asserted before returning to IDLE   
      when OHCRAP =>
        if start_test = '0' then
          next_state  <= IDLE;
        else
          next_state  <= OHCRAP;
        end if;
               
      -- keep sending data from the source for 1 ms to see that the 
      -- sink can shut down properly  
      when ENDRUN1 =>
        if maxendcnt = '1' then
          next_state  <= ENDRUN;
        else
          next_state  <= ENDRUN1;
        end if;
  
      -- make sure that start_test is de-asserted before returning to IDLE   
      when ENDRUN =>
        if start_test = '0' then
          next_state  <= IDLE;
        else
          next_state  <= ENDRUN;
        end if;
                
      when others =>
        next_state  <= IDLE;
      
    end case;    
  end process TEST_CONTROL_SM; 


-------------------------------------------------------------------------------
--  Other Output Assignments
-------------------------------------------------------------------------------
 
  test_active           <= run_enable;
  test_failed           <= dbg_reg3_r(47); -- error_det latched version 
  test_ready            <= idle_flag;
  source_clk            <= dut_clk;
  source_areset         <= '1' when areset = '1' else
                           '1' when control_dut_rst_bit = '1' else
                           '0';  
  source_fifo_empty     <= source_fifo_empty_flag;
  sink_areset           <= '1' when areset = '1' else
                           '1' when control_dut_rst_bit = '1' else
                           '0'; 
  sink_sample_clk       <= sample_clk;
  sink_fifo_full        <= sink_fifo_full_flag;
  scl                   <= scl_i;
  sda_out               <= sda_out_i;
  sda_oe                <= sda_oe_i;
  sreset_out            <= control_dut_rst_bit;
  sink_idelay_ch0       <= idelay_d0_bfield;
  sink_idelay_ch1       <= idelay_d1_bfield;
  sink_idelay_ch2       <= idelay_d2_bfield;
  sink_idelay_ch3       <= idelay_d3_bfield;
  
  -- xadc fan control for Genesys2 
  xadc_fan_ctrl        <= fan_on; 

end rtl;
