-------------------------------------------------------------------------------
--  Copyright (c) 2016 Eclektek LLC. All rights reserved.
--
--
--  Developed by: William Carter
--
--  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and  
--  associated documentation files (the "Software"), to deal with the Software without restriction,  
--  including without limitation the rights to use, copy, modify, merge, publish, sublicense and/or    
--  distribute copies of the Software, and to permit persons to whom the Software is furnished to do so,  
--  subject to the following conditions:
--
--      Redistributions of source code must retain the above copyright notice, this list of conditions 
--      and the following disclaimers.
--      Redistributions in binary form must reproduce the above copyright notice, this list of conditions 
--      and the following disclaimers in the documentation and/or other materials provided with the distribution.
--      Neither the names of William Carter, Eclektek LLC, nor the names of its 
--      contributors may be used to endorse or promote products derived from this Software without 
--      specific prior written permission from Eclektek LLC.
--      This Software is provided free of charge and Eclektek LLC reserves the right to sell copies of the 
--      Software in it original form. Anyone sub-licensing this Software will provide it free of charge.
--
--  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT 
--  LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
--  IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
--  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
--  SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
-------------------------------------------------------------------------------
--  History :                                                                       
--                                                                                                                  
--  Date      By    Comments                                         
--  --------  ----  ------------------------------------------------------------
--  09/20/16  W.C.  Created                                     
--
-------------------------------------------------------------------------------
--  Project       : Differential PMOD Challenge                                                          
--  PCB           : Genesys 2
--  Entity Name   : GENESYS2_BiPHL_Test
--  File Name     : GENESYS2_BiPHL_Test.vhd
--  Design Parent : none                                                          
--                                                                             
-------------------------------------------------------------------------------
--  Dependencies:
--
--  GENESYS2_BiPHL_Test.vhd is uses the following components:
--    Testbed                 Testbed.vhd
--    GENESYS2_Si570_INF      GENESYS2_Si570_INF .nco
--    DataSource_SDR          DataSource_SDR.vhd
--    DataSink_SDR            DataSink_SDR.vhd
--
--  GENESYS2_BiPHL_Test.vhd is uses the following packages:
--    Package        none
--
-------------------------------------------------------------------------------
--  GENESYS2_BiPHL_Test.vhd
--
--  This entity is the toplevel file for the Differential PMOD Challenge. This version
--  targets the Genesys 2 board and is designed to use the HW-FMC-105-DEBUG board to supply
--  a programmable clock source.
--
--  This project is a testbed framework for testing out your interface design using PMOD
--  connectors having only 8 wires arranged as 4 differential pairs. I've tried to make the 
--  design structure easy to use with a variety of data source/sink interface implementations
--  and enforce a common interface to encourage the creation of useful designs.   
--
--  The testing is pretty straight-forward. 
--  Simulation Mode:
--    SW7 when high sets Simulation mode. This allows simulation testbenches to run shorter
--        tests ( 16 words as opposed to 1 Mword ) and avoid having to simulate user UART inputs.
--        SW7 can be used on hardware to replicate simulations   
--    BTNL when strobed high will initiate a test run. This button is not conditioned in any way
--    Test runs source and sink 16 32-bit words or until there is an error. **  
--  Hardware Mode:
--    The UART is the means for controlling test runs and reporting results. The UART is encapsulated in
--    the Si570_INF IP. 
--    Test runs source data until the sink has acquired 1048576 32-bit words error free or until 
--    there is an error.  
--
--  On the GENESYS 2 Board JA is the Data Source connector and JB is the Data Sink connector
--
--  Monitor PMOD: On the GENESYS2 JC outputs signals useful for analyzing your design and the overall
--    functionality of a build.
--    JC(1) is the current Si570 clock output divided by 16
--    JC(2) is a replication of the FMC_SCL signal   
--    JC(3) is a replication of the FMC_SDA signal
--    JC(4) is unused
--    JC(10:7) is a copy of the JB signals that go to you Data Sink component   
--
--  This file constains generate statements to allow the user to easily configure the design
--  for a variety of test scenarios. There are two places in this file where you can setup your
--  design:
--    In the Constant Declarations section are a number of USE_xxx constants that need values consistant
--    with your design. 
--     USE_DIFFERENTIAL_IO           
--     USE_GENESYS2_BOARD           
--     USE_IDELAY2E_ILOGIC_RESOURCES
--     USE_IDDR_ILOGIC_RESOURCES    
--     USE_SDR_INTERFACE            
--     USE_BiPHASE_LEVEL_INTERFACE
--    Only 1 INTERFACE type can be used for a test
--    Depending on whether or not the ILOGIC resources are used the data signals driven into the Data
--    Sink must be selected in the Concurrent Assignments section. I elected not to use generate statements
--    here as there are a variety of options that might be wanted and the VHDL generate statement is limited
--    in functionality.
--  
-------------------------------------------------------------------------------
--  Limitations
--
------------------------------------------------------------------------------
--  ToDo
--    - BTNL is really for simulation purposes where it's easy to constrain the strobe
--      characteristics. It is possible to use it to start a test run when SW7 is high
--      but BTNL should be conditioned for use with the board.
--    - It would be nice to have a tcl automating project and file names
--    - Please do something for that poor PWM fan....
--
-------------------------------------------------------------------------------
--  NOTES:
--    - PLEASE read the text files in the doc folder of this project
--    - Requires a HW-FMC-105-DEBUG board with a Si570 programmable clock
--    - Requires a cable assembly to connect the SourcePMOD to the SinkPMOD
--    - I used Vivado 2016_2 to generate configuration files but any version of Vivado or
--      ISE 14.7 should work with no or minimal effort.
--    - Uses the Si570_INF IP supplied in netlist .nco format. The nco source appropriate
--      for your boards is instantiated in the Testbed.vhd component. Please read Si570_INF_IP.txt
--    - TMDS_33 drivers are open-source and require 50 OHM pullup resistors to VCC3V3 to get
--      an assertion to logic high. These should be on the PMOD cable assembly or loopback connector.
--    - The toplevel files for the Genesys 2 and Nexys Video boards are mostly the same
--      with the following differences:
--      - input clock is 200 MHz LVDS for Genesys2 and 100 MHz LVCMOS33 for Nexys Video
--      - the Vadj voltage is set by JP6 on the Genesys2 and via IO on the Nexys Video
--      - the Genesys2 has dedicated SCL/SDA lines for the FMC connector the Nexys Video shares
--        these with other devices on the board
--    - I have avoided using IP as much as is reasonable. I decided to use the Xilinx IP Clock Wizard
--      to produce the PLL IP as it is stable across Vivado/ISE versions and complicated enough to warrant
--      its use. I supply my versions of these in .xci source. They are easy enough to modify in Vivado.
--    - In recent version of Vivado .nco files are added to a project as Design Sources, .xci files are
--      added as IP Sources 
--
-------------------------------------------------------------------------------
--  CAUTIONS!!!
--
--    - The differential PMOD connectors on the Nexys Video and Genesys2 have NO protection for
--      the IO pins. You can seriously damage your board by not paying careful attention to how you
--      use them.
--    - Make sure that JP6 is set properly for Vadj = 2.5V
--
-- 
-------------------------------------------------------------------------------

Library ieee;
  use ieee.std_logic_1164.all;
  use ieee.std_logic_arith.all;
  use ieee.std_logic_unsigned.all;

library UNISIM;
  use UNISIM.vcomponents.all;
    
entity GENESYS2_BiPHL_Test is
  port(
        -- GLOBAL SIGNALS
        sysclk_p                        : in    std_logic; -- global external 200 MHz LVDS clock        
        sysclk_n                        : in    std_logic; -- global external 200 MHz LVDS clock        
        cpu_resetn                      : in    std_logic; -- normally '1', '0' when Reset button pressed 
        
        -- General I/O resources
        btnc                            : in    std_logic; -- Vadj HR Bank 17 normally '0', '1' when pressed
        btnd                            : in    std_logic; -- Vadj HR Bank 16 normally '0', '1' when pressed
        btnl                            : in    std_logic; -- Vadj HR Bank 15 normally '0', '1' when pressed
        btnr                            : in    std_logic; -- Vadj HR Bank 17 normally '0', '1' when pressed
        btnu                            : in    std_logic; -- Vadj HR Bank 17 normally '0', '1' when pressed
        sw                              : in    std_logic_vector(7 downto 0); -- Vadj HR Bank 15,16,17 or VCC3V3 Bank 14
        led                             : out   std_logic_vector(7 downto 0); -- VCC3V3 HR bank 14
        
        -- PMOD SourcePMOD Data Source
        SourcePMOD                      : out   std_logic_vector(7 downto 0); -- VCC3V3 HR Bank 14
        
        -- PMOD SinkPMOD Data Sink
        SinkPMOD                        : in    std_logic_vector(7 downto 0); -- VCC3V3 HR Bank 14 
        
        -- PMOD MonitorPMOD signal monitor outputs for debugging
        MonitorPMOD                     : out   std_logic_vector(7 downto 0); -- VCC3V3 HR Bank 13
        
        -- I2C PHYSICAL INTERFACE
        FMC_SCL                         : out   std_logic; -- VCC3V3 Bank 12
        FMC_SDA                         : inout std_logic; -- VCC3V3 Bank 12
        
        -- UART
        uart_rx_out                     : out   std_logic; -- VCC3V3 Bank 12
        uart_tx_in                      : in    std_logic; -- VCC3V3 Bank 12
        
        -- Genesys2 FAN
        FAN_PWM                         : out   std_logic; -- VCC3V3 Bank 14 
        
        -- FMC 
        fmc_clk0_m2c_n                  : in    std_logic; -- Vadj Bank 17
        fmc_clk0_m2c_p                  : in    std_logic  -- Vadj Bank 17
  );
end entity;

architecture rtl of GENESYS2_BiPHL_Test is

-------------------------------------------------------------------------------
--  Component Declarations
-------------------------------------------------------------------------------

  component Testbed
    generic(
      IS_GENESYS2_BOARD        : natural range 1 downto 0 := 1);                -- 1 when targetting the Genesys 2 board
    port(
        -- Testbed interface
        tb_clock_100                    : in    std_logic;                      -- 100 MHz testbed clock
        dut_clk                         : in    std_logic;                      -- programmable DUT Data Source clock
        sample_clk                      : in    std_logic;                      -- Data Sink sample clock unrelated to dut_clk
        areset                          : in    std_logic;                      -- asynchronous reset
        sreset_out                      : out   std_logic;                      -- synchronous reset from UART control 
                                                                                        
        sim_run_words_m1                : in    std_logic_vector(23 downto 0);  -- number of 32-bit words minus 1 in a test run 
        run_words_m1                    : in    std_logic_vector(23 downto 0);  -- number of 32-bit words minus 1 in a test run 
        sim_source_lfsr_seed            : in    std_logic_vector(31 downto 0);  -- anything that ISN'T X"00000000"
        source_lfsr_seed                : in    std_logic_vector(31 downto 0);  -- anything that ISN'T X"00000000"
        test_id                         : in    std_logic_vector(47 downto 0);    
        
        -- Board User Switch inputs
        simulation_mode_flag            : in    std_logic;                      -- '1' sets simulation mode, '0' sets hardware mode
                
        -- Board User Button inputs
        sim_start_stb                   : in    std_logic;                      -- asserted high to start a simulation test run
        
        
        -- LED suitable indicators
        test_active                     : out   std_logic;                      
        test_failed                     : out   std_logic;  
        test_ready                      : out   std_logic;                    
        tb_clock_100_heartbeat          : out   std_logic;                    
        dut_clk_heartbeat               : out   std_logic;    
        xadc_fan_ctrl                   : out   std_logic;    
        
        -- PMOD suitable monitor signals                
        si570_div16_monitor             : out   std_logic;
        
        -- I2C PHYSICAL INTERFACE to operate the FMC Si570 programmable clock
        scl                             : out   std_logic; 
        sda_in                          : in    std_logic; 
        sda_out                         : out   std_logic; 
        sda_oe                          : out   std_logic; 
        
        -- UART
        uart_rx_out                     : out   std_logic;
        uart_tx_in                      : in    std_logic;
        
        -- IDELAY controls
        sink_idelay_ch0                 : out   std_logic_vector(4 downto 0);
        sink_idelay_ch1                 : out   std_logic_vector(4 downto 0);
        sink_idelay_ch2                 : out   std_logic_vector(4 downto 0);
        sink_idelay_ch3                 : out   std_logic_vector(4 downto 0);
                        
        -- Data Source DUT Interface                                                
        source_clk                      : out   std_logic;                      -- source PMOD data bit clock
        source_areset                   : out   std_logic;                      -- source asynchronous reset input
        source_fifo_data                : out   std_logic_vector(31 downto 0);  -- data to be transmitted
        source_fifo_pdata               : out   std_logic_vector(3 downto 0);   -- source control from testbed
        source_fifo_empty               : out   std_logic;                      -- '1' indiates data available
        source_fifo_rden                : in    std_logic;                      -- fifo read enable        
        source_data_out                 : in    std_logic_vector(3 downto 0);   -- to output PMOD    
        
        -- Data Sink DUT  Interface
        sink_areset                     : out   std_logic;                      -- asynchronous reset
        sink_sample_clk                 : out   std_logic;                      -- sample clock for SinkPMOD signals
        sink_fifo_full                  : out   std_logic;                      -- '0' if FIFO can receive data
        sink_fifo_data                  : in    std_logic_vector(31 downto 0);  -- data received
        sink_fifo_pdata                 : in    std_logic_vector(3 downto 0);   -- sink control to testbed
        sink_fifo_wren                  : in    std_logic                       -- fifo write enable        
    );
  end component;
  
  
component DataSource_SDR
  port(
        -- GLOBAL SIGNALS
        tclk                            : in    std_logic; -- variable test clock
        areset                          : in    std_logic; -- asynchronous reset
        
        -- interface  
        data_in                         : in    std_logic_vector(31 downto 0); -- data to be transmitted
        pdata_in                        : in    std_logic_vector(3 downto 0);  -- control from testbed
        fifo_empty                      : in    std_logic;                     -- '1' indiates data available
        fifo_rden                       : out   std_logic;                     -- fifo read enable        
                
        -- PMOD 3 bits data plus clock
        data_out                        : out   std_logic_vector(3 downto 0)        
  );
end component;
        
       
component DataSink_SDR 
  port(
        -- GLOBAL SIGNALS
        fs_clk                          : in    std_logic;  
        areset                          : in    std_logic; -- asynchronous reset
                
        -- PMOD 3 bits data plus clock
        data_in                         : in    std_logic_vector(3 downto 0);        
        
        -- Received data
        data_out                        : out   std_logic_vector(31 downto 0); -- data received
        pdata_out                       : out   std_logic_vector(3 downto 0);  -- control to testbed
        fifo_full                       : in    std_logic;                     -- '0' if FIFO can receive data
        fifo_wren                       : out   std_logic;                     -- fifo write enable
        
        -- debug
        d                               : out   std_logic_vector(31 downto 0); -- data received
        d_valid                         : out   std_logic        
  );
end component;
  
component DataSource_BiPH_L
  port(
        -- GLOBAL SIGNALS
        tclk                            : in    std_logic; -- variable test clock
        areset                          : in    std_logic; -- asynchronous reset
        
        -- interface  
        data_in                         : in    std_logic_vector(31 downto 0); -- data to be transmitted
        pdata_in                        : in    std_logic_vector(3 downto 0);  -- control from testbed
        fifo_empty                      : in    std_logic;                     -- '1' indiates data available
        fifo_rden                       : out   std_logic;                     -- fifo read enable        
                
        -- PMOD 4 bits data
        data_out                        : out   std_logic_vector(3 downto 0)        
  );
end component;
        
       
component DataSink_BiPH_L 
  port(
        -- GLOBAL SIGNALS
        fs_clk                          : in    std_logic;  
        areset                          : in    std_logic; -- asynchronous reset
                
        -- PMOD 4 bits data
        data_in                         : in    std_logic_vector(3 downto 0);        
        
        -- Received data
        data_out                        : out   std_logic_vector(31 downto 0); -- data received
        pdata_out                       : out   std_logic_vector(3 downto 0);  -- control to testbed
        fifo_full                       : in    std_logic;                     -- '0' if FIFO can receive data
        fifo_wren                       : out   std_logic;                     -- fifo write enable
        
        -- debug
        d                               : out   std_logic_vector(31 downto 0); -- data received
        d_valid                         : out   std_logic        
  );
end component;

component clk_wiz_0
port
 (-- Clock in ports
  clk_in1_p         : in     std_logic;
  clk_in1_n         : in     std_logic;
  -- Clock out ports
  clk_out1          : out    std_logic;
  clk_out2          : out    std_logic;
  clk_out3          : out    std_logic;
  -- Status and control signals
  reset             : in     std_logic;
  locked            : out    std_logic
 );
end component;



-------------------------------------------------------------------------------
--  Type Declarations
-------------------------------------------------------------------------------

type idelay_count  is array (3 downto 0) of std_logic_vector(4 downto 0);

-------------------------------------------------------------------------------
--  Constant Declarations
-------------------------------------------------------------------------------
 
-- Set these USE_XXXX constants for your board, Data Interface IOSTANDARD, and ILOGIC choices
-- NOTE: Your .xdc constraints file MUST support the selected options
constant USE_DIFFERENTIAL_IO                      : natural range 1 downto 0 := 1; -- SourcePMOD/SinkPMOD use TMDS_33 when 1
constant USE_GENESYS2_BOARD                       : natural range 1 downto 0 := 1; -- Target the Genesys 2 board when 1
constant USE_IDELAY2E_ILOGIC_RESOURCES            : natural range 1 downto 0 := 0; -- use IDELAY2E ILOGIC when 1
constant USE_IDDR_ILOGIC_RESOURCES                : natural range 1 downto 0 := 0; -- use IDDR ILOGIC when 1

  -- set one, but only one, of the USE_XXX_INTERFACE constants to 1
constant USE_SDR_INTERFACE                        : natural range 1 downto 0 := 0;
constant USE_BiPHASE_LEVEL_INTERFACE              : natural range 1 downto 0 := 1;
-- Set these USE_XXXX constants for your board, Data Interface IOSTANDARD, and ILOGIC choices

constant TEST_VERSION                             : std_logic_vector(15 downto 0) := X"0001"; --  
constant SIM_TEST_COUNT_END_M1                    : std_logic_vector(23 downto 0) := X"00000F"; 
constant REAL_TEST_COUNT_END_M1                   : std_logic_vector(23 downto 0) := X"0FFFFF"; 


-------------------------------------------------------------------------------
--  Signal Declarations
-------------------------------------------------------------------------------

signal clk100                     : std_logic;
signal clk200                     : std_logic;
signal clk_fs                     : std_logic;
signal areset                     : std_logic;
signal sreset                     : std_logic;
signal idlyctrl_rst               : std_logic;
signal sda_in_i                   : std_logic;
signal sda_out_i                  : std_logic;
signal scl_i                      : std_logic;
signal sda_oe_i                   : std_logic;
signal iclk100_heartbeat_led      : std_logic;
signal iclk_si570_heartbeat_led   : std_logic;
signal vclk                       : std_logic;
signal vclk_n                     : std_logic;
signal vcnt                       : std_logic_vector(31 downto 0);
signal start_test                 : std_logic;                                             
signal is_simulation              : std_logic;
signal pmod_dsource_p             : std_logic_vector(3 downto 0);
signal pmod_dsource_n             : std_logic_vector(3 downto 0);
signal pmod_dsink_p               : std_logic_vector(3 downto 0);
signal pmod_dsink_n               : std_logic_vector(3 downto 0);
signal sim_start                  : std_logic;
signal test_active_led            : std_logic;
signal test_failed_led            : std_logic;
signal test_ready_led             : std_logic;
signal tb_clock_100_heartbeat_led : std_logic;
signal dut_clk_heartbeat_led      : std_logic;
signal si570_div16                : std_logic;
signal isink_sample_clk           : std_logic;
signal sink_data_in               : std_logic_vector(3 downto 0);   
signal sink_data_in_r             : std_logic_vector(3 downto 0);   
signal sink_data_in_r1            : std_logic_vector(3 downto 0);   
signal sink_data_in_r2            : std_logic_vector(3 downto 0);   
signal sink_data_in_r3            : std_logic_vector(3 downto 0);   
signal sink_data_in_r4            : std_logic_vector(3 downto 0);   
signal sink_data_in_r5            : std_logic_vector(3 downto 0);   
signal sink_data_in_r6            : std_logic_vector(3 downto 0);   
signal sink_data_in_dly           : std_logic_vector(3 downto 0);
signal sink_datain                : std_logic_vector(3 downto 0);   
signal sink_datain_re             : std_logic_vector(3 downto 0);   
signal sink_datain_fe             : std_logic_vector(3 downto 0);   
signal source_data_out            : std_logic_vector(3 downto 0);   
signal source_arst                : std_logic;
signal sink_arst                  : std_logic;
signal source_clock               : std_logic;
signal source_din                 : std_logic_vector(31 downto 0);
signal source_pdin                : std_logic_vector(3 downto 0);
signal source_data_available_n    : std_logic;
signal source_rden                : std_logic;
signal sink_dout                  : std_logic_vector(31 downto 0);
signal sink_pdout                 : std_logic_vector(3 downto 0);
signal sink_fifo_full_flg         : std_logic;
signal sink_data_wr               : std_logic;
signal pll_is_locked              : std_logic;
signal idelay_count_in            : idelay_count;
signal idelay_count_out           : idelay_count;
signal idelay_ld                  : std_logic_vector(3 downto 0);
signal isink_idelay_ch0           : std_logic_vector(4 downto 0);
signal isink_idelay_ch1           : std_logic_vector(4 downto 0);
signal isink_idelay_ch2           : std_logic_vector(4 downto 0);
signal isink_idelay_ch3           : std_logic_vector(4 downto 0);
signal dcnt                       : std_logic_vector(19 downto 0);
signal test_setup_id              : std_logic_vector(47 downto 0);
signal ixadc_fan_ctrl             : std_logic;

signal sink_debug_data            : std_logic_vector(31 downto 0);
signal sink_debug_dvalid          : std_logic;

begin


-------------------------------------------------------------------------------
--  Component Instantiations
-------------------------------------------------------------------------------

    TESTBED_inst : Testbed
    generic map(
      IS_GENESYS2_BOARD   => USE_GENESYS2_BOARD)           -- 1 when targetting the Genesys 2 board
    port map(                             
          tb_clock_100                    => clk100,
          dut_clk                         => vclk,
          sample_clk                      => clk_fs,
          areset                          => areset,
          sreset_out                      => sreset,
          sim_run_words_m1                => SIM_TEST_COUNT_END_M1,
          run_words_m1                    => REAL_TEST_COUNT_END_M1,
          sim_source_lfsr_seed            => "11110000111100001111000011110000", -- 0xF0F0F0F0
          source_lfsr_seed                => "11110000111100001111000011110000",
          test_id                         => test_setup_id,
          simulation_mode_flag            => is_simulation,
          sim_start_stb                   => sim_start,
          test_active                     => test_active_led,
          test_failed                     => test_failed_led,
          test_ready                      => test_ready_led,
          tb_clock_100_heartbeat          => tb_clock_100_heartbeat_led,
          dut_clk_heartbeat               => dut_clk_heartbeat_led,
          xadc_fan_ctrl                   => ixadc_fan_ctrl,
          si570_div16_monitor             => si570_div16,
          scl                             => scl_i,
          sda_in                          => sda_in_i,
          sda_out                         => sda_out_i,
          sda_oe                          => sda_oe_i,
          uart_rx_out                     => uart_rx_out,
          uart_tx_in                      => uart_tx_in,
          sink_idelay_ch0                 => isink_idelay_ch0,
          sink_idelay_ch1                 => isink_idelay_ch1,
          sink_idelay_ch2                 => isink_idelay_ch2,
          sink_idelay_ch3                 => isink_idelay_ch3,
          source_clk                      => source_clock,
          source_areset                   => source_arst,
          source_fifo_data                => source_din,
          source_fifo_pdata               => source_pdin,
          source_fifo_empty               => source_data_available_n,
          source_fifo_rden                => source_rden,
          source_data_out                 => source_data_out,
          sink_areset                     => sink_arst,
          sink_sample_clk                 => isink_sample_clk,
          sink_fifo_full                  => sink_fifo_full_flg,
          sink_fifo_data                  => sink_dout,         
          sink_fifo_pdata                 => sink_pdout,        
          sink_fifo_wren                  => sink_data_wr);
    
  GEN_inf0 : if USE_BiPHASE_LEVEL_INTERFACE = 1 generate
    SRC_inst : DataSource_BiPH_L
    port map(                                                             
          tclk                       => source_clock,                
          areset                     => source_arst,         
          data_in                    => source_din,                                 
          pdata_in                   => source_pdin,                                
          fifo_empty                 => source_data_available_n,                    
          fifo_rden                  => source_rden,                
          data_out                   => source_data_out);
    
          
    SINK_inst : DataSink_BiPH_L
    port map(
          fs_clk                     => isink_sample_clk,
          areset                     => sink_arst,
          data_in                    => sink_datain, 
          data_out                   => sink_dout,
          pdata_out                  => sink_pdout,
          fifo_full                  => sink_fifo_full_flg,
          fifo_wren                  => sink_data_wr,
          d                          => sink_debug_data,
          d_valid                    => sink_debug_dvalid);
  end generate;
    
  GEN_inf1 : if USE_SDR_INTERFACE = 1 generate
    SRC_inst : DataSource_SDR
    port map(
                                                             
          -- PMOD 3 bits data plus clock                     
          tclk                       => source_clock,                
          areset                     => source_arst,         
          data_in                    => source_din,                                 
          pdata_in                   => source_pdin,                                
          fifo_empty                 => source_data_available_n,                    
          fifo_rden                  => source_rden,                
          data_out                   => source_data_out);
    
          
    SINK_inst : DataSink_SDR
    port map(
          fs_clk                     => isink_sample_clk,
          areset                     => sink_arst,
          data_in                    => sink_datain,
          data_out                   => sink_dout,
          pdata_out                  => sink_pdout,
          fifo_full                  => sink_fifo_full_flg,
          fifo_wren                  => sink_data_wr,
          d                          => sink_debug_data,
          d_valid                    => sink_debug_dvalid);
  end generate;
                

  -- Si570 programmable clock input from FMC connector
  Si570clk_inst : IBUFGDS_DIFF_OUT
  generic map (
    DIFF_TERM     => FALSE, -- Differential Termination
    IBUF_LOW_PWR  => FALSE, -- Low power (TRUE) vs. performance (FALSE) setting for referenced I/O standards
    IOSTANDARD    => "LVDS_25") -- Specify the input I/O standard
  port map (
    O     => vclk, -- Buffer diff_p output
    OB    => vclk_n, -- Buffer diff_n output
    I     => fmc_clk0_m2c_p, -- Diff_p buffer input (connect directly to top-level port)
    IB    => fmc_clk0_m2c_n -- Diff_n buffer input (connect directly to top-level port)
  );
  
  PLL0_inst : clk_wiz_0
   port map ( 
   clk_in1_p  => sysclk_p,
   clk_in1_n  => sysclk_n,
   clk_out1   => clk200,
   clk_out2   => clk100,
   clk_out3   => clk_fs,
   reset      => '0',
   locked     => pll_is_locked            
  );
  
  
  -- SourcePMOD output buffers
  -- If using eight PMOD LVCMOS33 IO you need to modify the following to generate 8 OBUFs and
  -- connect all *_n and *_p pins  
  GEN_obuf : for i in 3 downto 0  generate
    -- SourcePMOD LVCMOS33 output buffers
    -- OBUF: Single-ended Output Buffer
    -- 7 Series
    -- Xilinx HDL Libraries Guide, version 13.4   
    GEN_obufs : if USE_DIFFERENTIAL_IO = 0 generate
      OBUF_inst : OBUF
      generic map (
        DRIVE       => 16,
        IOSTANDARD  => "LVCMOS33",
        SLEW        => "FAST")
      port map (
        O => pmod_dsource_p(i), -- Buffer output (connect directly to top-level port)
        I => source_data_out(i) -- Buffer input
      );  
    end generate;
    -- SourcePMOD LVCMOS33 output buffers
    -- OBUFDS: Differential Output Buffer
    -- 7 Series
    -- Xilinx HDL Libraries Guide, version 13.4   
    GEN_obufd : if USE_DIFFERENTIAL_IO = 1 generate
      OBUFDS_inst : OBUFDS
      generic map (
        IOSTANDARD  => "TMDS_33", -- Specify the output I/O standard
        SLEW        => "FAST")    -- Specify the output slew rate as "SLOW" or "FAST"
      port map (
        O     => pmod_dsource_p(i),   -- Diff_p output (connect directly to top-level port)
        OB    => pmod_dsource_n(i),   -- Diff_n output (connect directly to top-level port)
        I     => source_data_out(i)   -- Buffer input
      );
    end generate;    
  end generate;

  -- SinkPMOD input buffers
  -- If using eight PMOD LVCMOS33 IO you need to modify the following to generate 8 IBUFs and
  -- connect all *_n and *_p pins  
  GEN_ibuf : for i in 3 downto 0 generate
    GEN_ibufs : if USE_DIFFERENTIAL_IO = 0 generate
      -- IBUF: Single-ended Input Buffer
      -- 7 Series
      -- Xilinx HDL Libraries Guide, version 13.4
      IBUF_inst : IBUF
      generic map (
        IBUF_LOW_PWR  => FALSE, -- Low power (TRUE) vs. performance (FALSE) setting for referenced I/O standards
        IOSTANDARD    => "LVCMOS33")
      port map (
        O => sink_data_in(i), -- Buffer output
        I => pmod_dsink_p(i) -- Buffer input (connect directly to top-level port)
      );
      end generate;
    GEN_ibufd : if USE_DIFFERENTIAL_IO = 1 generate
      -- IBUFDS: Differential Input Buffer
      -- 7 Series
      -- Xilinx HDL Libraries Guide, version 13.4
    IBUFDS_inst : IBUFDS
    generic map (
      DIFF_TERM     => TRUE, -- Differential Termination
      IBUF_LOW_PWR  => TRUE,  -- Low power (TRUE) vs. performance (FALSE) setting for referenced I/O standards
      IOSTANDARD    => "TMDS_33")
    port map (
      O   => sink_data_in(i), -- Buffer output
      I   => pmod_dsink_p(i), -- Diff_p buffer input (connect directly to top-level port)
      IB  => pmod_dsink_n(i)  -- Diff_n buffer input (connect directly to top-level port)
    );
    end generate;
  end generate;
 
  
     
  -- IDELAYCTRL: IDELAYE2/ODELAYE2 Tap Delay Value Control
  -- 7 Series
  -- Xilinx HDL Libraries Guide, version 13.4  
  GEN_idlyctl : for i in 3 downto 0 generate
    GEN_idlyctl1 : if USE_IDELAY2E_ILOGIC_RESOURCES = 0 generate
      IDELAYCTRL_inst : IDELAYCTRL
      port map (
        RDY     => open, -- 1-bit output: Ready output
        REFCLK  => clk200, -- 1-bit input: Reference clock input
        RST     => idlyctrl_rst -- 1-bit input: Active high reset input
      );
    end generate;
  end generate;
 
  -- NOTE: according to UG471 IDELAYCTRL must be reset after configuration to ensure proper
  --       IDELAY and ODELAY operation

  
  
  --  IDELAYs for the SinkPMOD inputs
  GEN_idelay : for i in 3 downto 0 generate
    GEN_idelay1 : if USE_IDELAY2E_ILOGIC_RESOURCES = 1 generate
      -- IDELAYE2: Input Fixed or Variable Delay Element
      -- 7 Series
      -- Xilinx HDL Libraries Guide, version 13.4
      IDELAYE2_inst : IDELAYE2
      generic map (
        CINVCTRL_SEL            => "FALSE",     -- Enable dynamic clock inversion (FALSE, TRUE)
        DELAY_SRC               => "DATAIN",   -- Delay input (IDATAIN, DATAIN)
        HIGH_PERFORMANCE_MODE   => "TRUE",     -- Reduced jitter ("TRUE"), Reduced power ("FALSE")
        IDELAY_TYPE             => "VAR_LOAD",  -- FIXED, VARIABLE, VAR_LOAD, VAR_LOAD_PIPE
        IDELAY_VALUE            => 0,          -- Input delay tap setting (0-31)
        PIPE_SEL                => "FALSE",     -- Select pipelined mode, FALSE, TRUE
        REFCLK_FREQUENCY        => 200.0,       -- IDELAYCTRL clock input frequency in MHz (190.0-210.0).
        SIGNAL_PATTERN          => "DATA"       -- DATA, CLOCK input signal
      )
      port map (
        CNTVALUEOUT       => idelay_count_out(i), -- 5-bit output: Counter value output
        DATAOUT           => sink_data_in_dly(i), -- 1-bit output: Delayed data output
        C                 => clk_fs, -- 1-bit input: Clock input
        CE                => '0', -- 1-bit input: Active high enable increment/decrement input
        CINVCTRL          => '0', -- 1-bit input: Dynamic clock inversion input
        CNTVALUEIN        => idelay_count_in(i), -- 5-bit input: Counter value input
        DATAIN            => sink_data_in_r2(i), -- 1-bit input: Internal delay data input
        IDATAIN           => '0', -- 1-bit input: Data input from the I/O
        INC               => '0', -- 1-bit input: Increment / Decrement tap delay input
        LD                => idelay_ld(i), -- 1-bit input: Load IDELAY_VALUE input
        LDPIPEEN          => '0', -- 1-bit input: Enable PIPELINE register to load data input
        REGRST            => '0' -- 1-bit input: Active-high reset tap-delay input
      );
    end generate;
  end generate;
 
  
  GEN_iddr : for i in 3 downto 0 generate
    GEN_iddr1 : if USE_IDDR_ILOGIC_RESOURCES = 1 generate
      -- IDDR: Double Data Rate Input Register with Set, Reset
      -- and Clock Enable.
      -- 7 Series
      -- Xilinx HDL Libraries Guide, version 13.4
      IDDR_inst : IDDR
      generic map (
        DDR_CLK_EDGE    => "OPPOSITE_EDGE", -- "OPPOSITE_EDGE", "SAME_EDGE"  or "SAME_EDGE_PIPELINED"
        INIT_Q1         => '0', -- Initial value of Q1: '0' or '1'
        INIT_Q2         => '0', -- Initial value of Q2: '0' or '1'
        SRTYPE          => "SYNC") -- Set/Reset type: "SYNC" or "ASYNC"
      port map (
        Q1  => sink_datain_re(i), -- 1-bit output for positive edge of clock
        Q2  => sink_datain_fe(i), -- 1-bit output for negative edge of clock
        C   => clk_fs, -- 1-bit clock input
        CE  => '1', -- 1-bit clock enable input
        D   => sink_data_in_dly(i), -- 1-bit DDR data input
        R   => '0', -- 1-bit reset
        S   => '0' -- 1-bit set
      );    
    end generate;
  end generate;
                 
-------------------------------------------------------------------------------
--  Concurrent Assignments
-------------------------------------------------------------------------------

  areset  <= not cpu_resetn; 
  idlyctrl_rst  <= sreset or areset;
  
   
  sda_in_i  <= FMC_SDA;
  
  -- assign functionality to the user switches
  is_simulation     <= '1' when sw(7) = '1' else '0';  
  
  -- assign functionality to the user buttons
  sim_start         <= btnl; -- condition in case user wants to use buttons
  
  -- fill out information for Reg1 test_id
  -- board and configuration options
  test_setup_id(47)             <= '1' when USE_GENESYS2_BOARD = 1 else '0';
  test_setup_id(46)             <= '1' when USE_DIFFERENTIAL_IO = 1 else '0';
  test_setup_id(45)             <= '1' when USE_IDELAY2E_ILOGIC_RESOURCES = 1 else '0';
  test_setup_id(44)             <= '1' when USE_IDDR_ILOGIC_RESOURCES = 1 else '0';
  test_setup_id(43 downto 40)   <= X"0"; -- available for expansion
  
  -- interface options
  test_setup_id(39 downto 32)   <= X"00" when USE_SDR_INTERFACE = 1 else
                                   X"01" when USE_BiPHASE_LEVEL_INTERFACE = 1 else 
                                   X"00";
  test_setup_id(31 downto 16)   <= X"0000";       -- reserved  
  test_setup_id(15 downto 0)    <= TEST_VERSION;  -- Differential PMOD Challenge Version ID 
    
  -- Testbed PMOD Data Sink connections
  pmod_dsink_n(0)     <= SinkPMOD(0); -- SinkPMOD(1)    
  pmod_dsink_p(0)     <= SinkPMOD(1); -- SinkPMOD(2) 
  pmod_dsink_n(1)     <= SinkPMOD(2); -- SinkPMOD(3) 
  pmod_dsink_p(1)     <= SinkPMOD(3); -- SinkPMOD(4) 
  pmod_dsink_n(2)     <= SinkPMOD(4); -- SinkPMOD(7) 
  pmod_dsink_p(2)     <= SinkPMOD(5); -- SinkPMOD(8) 
  pmod_dsink_n(3)     <= SinkPMOD(6); -- SinkPMOD(9) 
  pmod_dsink_p(3)     <= SinkPMOD(7); -- SinkPMOD(10)
 
  
  -- Input schematic
  -- SinkPMOD -- pmod_dsink_p --> | IBUF or | -- sink_data_in --> | CONDITIONING | -- sink_data_in_rX -->
  --          -- pmod_dsink_n --> | IBUFDS  |          clk_fs --> |              |
  --
  --   -- sink_data_in_rX --> | IDELAY2E OPTION | -- sink_data_in_dly -->
  --               clk_fs --> |                 |
  --
  --   -- sink_data_in_dly --> | IDDR OPTION | -- sink_datain_re -->
  --                clk_fs --> |             | -- sink_datain_fe -->
  --
  --     -- sink_datain_re --> | SELECT | 
  --     -- sink_datain_fe --> |  ONE   | -- sink_datain --> [ SINK_inst ]
  --    -- sink_data_in_rX --> |        |
  --   -- sink_data_in_dly --> |        |
  
  
  -- Options for driving the Data Sink
  -- Uncomment one depending on your design
  -- This selection is NOT covered by the constant OPTION settings
--  sink_datain   <= sink_datain_re when sw(1) = '1' else sink_datain_fe; 
--  sink_datain   <= sink_datain_re; 
--  sink_datain   <= sink_datain_fe; 
  sink_datain   <= sink_data_in_r2; 
  -- Uncomment one depending on your design
                     
-------------------------------------------------------------------------------
--  Process Declarations
-------------------------------------------------------------------------------
   
  -- SinkPMOD data sampler                   
  -- Here is where you might want to try applying a bit of signal
  -- conditioning                   
  process(clk_fs)
  begin
    if rising_edge(clk_fs) then
      sink_data_in_r  <= sink_data_in;
      sink_data_in_r1 <= sink_data_in_r;
      sink_data_in_r2 <= sink_data_in_r1;
      sink_data_in_r3 <= sink_data_in_r2;
    end if;  
  end process;   
                     
  GEN_idly : if USE_IDELAY2E_ILOGIC_RESOURCES = 1 generate
    
    -- hook these up to a UART control Register
    idelay_count_in(3)    <= isink_idelay_ch3; 
    idelay_count_in(2)    <= isink_idelay_ch2; 
    idelay_count_in(1)    <= isink_idelay_ch1; 
    idelay_count_in(0)    <= isink_idelay_ch0; 
  
    process(clk_fs)
    begin
      if rising_edge(clk_fs) then
        dcnt  <= dcnt + 1;
      end if;  
    end process;   
                     
    process(clk_fs)
    begin
      if rising_edge(clk_fs) then
        dcnt  <= dcnt + 1;
      end if;  
    end process;   
    
    process(clk_fs)
    begin
      if rising_edge(clk_fs) then
        if dcnt = X"00000" and idelay_count_in(2) /= idelay_count_out(2) then
          idelay_ld(2) <= '1';
        else
          idelay_ld(2) <= '0';
        end if;
      end if;       
    end process;   
    
    process(clk_fs)
    begin
      if rising_edge(clk_fs) then
        if dcnt = X"00000" and idelay_count_in(1) /= idelay_count_out(1) then
          idelay_ld(1) <= '1';
        else
          idelay_ld(1) <= '0';
        end if;
      end if;       
    end process;   
    
    process(clk_fs)
    begin
      if rising_edge(clk_fs) then
        if dcnt = X"00000" and idelay_count_in(0) /= idelay_count_out(0) then
          idelay_ld(0) <= '1';
        else
          idelay_ld(0) <= '0';
        end if;
      end if;       
    end process;   
  end generate;
    
  
-------------------------------------------------------------------------------
--  Other Output Assignments
-------------------------------------------------------------------------------
                     
  led(7)          <= tb_clock_100_heartbeat_led;
  led(6)          <= dut_clk_heartbeat_led; 
  led(5)          <= is_simulation;           
  led(4)          <= test_active_led;           
  led(3)          <= test_ready_led;
  led(2)          <= '0';           
  led(1)          <= sw(1);           
  led(0)          <= test_failed_led;
    
  FMC_SDA         <= sda_out_i when sda_oe_i = '1' else 'Z';
  FMC_SCL         <= scl_i;
  
  -- PMOD JA is used as a status monitor
  MonitorPMOD(0)       <= si570_div16;   -- MonitorPMOD(2) unused
  MonitorPMOD(1)       <= scl_i;         -- MonitorPMOD(3) will show SCL
  MonitorPMOD(2)       <= sda_in_i;      -- MonitorPMOD(4) will show SDA
  MonitorPMOD(3)       <= 'Z';           -- MonitorPMOD(7) unused
  MonitorPMOD(7 downto 4)  <= sink_data_in_r3;
  
  -- Testbed PMOD Data Source connections
  -- If LVCMOS33 allow the 50 ohm pullups to set *_n output states when 50 ohm pullups are 
  -- used on the physical loopback connector. If you want to try using all eight lines modify
  -- as desired and 
  GEN_jsourceout : if USE_DIFFERENTIAL_IO = 0 generate
    SourcePMOD(0)       <= 'Z'; -- SourcePMOD(1) -- DON'T FIGHT THE 50 ohm PULL-UP  
    SourcePMOD(1)       <= pmod_dsource_p(0); -- SourcePMOD(2)
    SourcePMOD(2)       <= 'Z'; -- SourcePMOD(3)
    SourcePMOD(3)       <= pmod_dsource_p(1); -- SourcePMOD(4)
    SourcePMOD(4)       <= 'Z'; -- SourcePMOD(7)
    SourcePMOD(5)       <= pmod_dsource_p(2); -- SourcePMOD(8)
    SourcePMOD(6)       <= 'Z'; -- SourcePMOD(9)
    SourcePMOD(7)       <= pmod_dsource_p(3); -- SourcePMOD(10)
  end generate;

  -- If TMDS_33 the physical loopback connector mus have 50 ohm pull-up terminators
  GEN_jsourceout1 : if USE_DIFFERENTIAL_IO = 1 generate
    SourcePMOD(0)       <= pmod_dsource_n(0); -- SourcePMOD(1)   
    SourcePMOD(1)       <= pmod_dsource_p(0); -- SourcePMOD(2)
    SourcePMOD(2)       <= pmod_dsource_n(1); -- SourcePMOD(3)
    SourcePMOD(3)       <= pmod_dsource_p(1); -- SourcePMOD(4)
    SourcePMOD(4)       <= pmod_dsource_n(2); -- SourcePMOD(7)
    SourcePMOD(5)       <= pmod_dsource_p(2); -- SourcePMOD(8)
    SourcePMOD(6)       <= pmod_dsource_n(3); -- SourcePMOD(9)
    SourcePMOD(7)       <= pmod_dsource_p(3); -- SourcePMOD(10)
  end generate;

  FAN_PWM               <= ixadc_fan_ctrl;
    
end rtl;
